import numpy
from numba import njit


#@njit(cache=True)
def enclosing_points_1d(
        xp: numpy.ndarray,
        x: numpy.ndarray,
        regular_xp=False,
        period: float = None
):
    """
    Find surrounding indices for value x[j] in vector xp such
    that  xp[i-1] <= x[j] < xp[i]

    To note for non-periodic sequences values outside of xp require special
    attention- specifically:
        for x < xp[0]:
            we have x[x<xp[0]] < xp[indices[1,:]], and indices[0,:] is
            undefined (clipped to 0).

        for x >= xp[-1]:
            we have x[x>=xp[-1]] >= xp[indices[0,:]], and indices[1,:] is
            undefined (clipped to nx).

    Parameters
    ----------
    :param x: 1-D array_like of length nx
            The x-coordinates at which to evaluate the interpolated values with
            nx entries.

    :param xp: 1-D sequence of floats
        The x-coordinates of the data points with nxp entries.

    :return: [2 by nx] numpy array of integer (dtype='int64') indices such that
        xp[indices[0,:]] <= x[:] < xp[indices[1,:]] with the exception of
        points x outside of the domain of xp.


    """

    x = numpy.atleast_1d(x)
    nxp = len(xp)
    nx = len(x)

    if xp[-1] < xp[0]:
        # make sure we are in a coordinate frame where the vector is
        # in ascending order.
        x =  xp[0] - x
        xp = xp[0] - xp

    if period is not None:
        x = (x - xp[0]) % period + xp[0]

    indices = numpy.empty((2, nx), dtype='int64')
    if not regular_xp:
        for ix, x_ix in enumerate(x):
            # find ixp such that  xp[ixp-1] <= x[ix] < xp[ixp]
            ixp = numpy.searchsorted(xp, x_ix, side='right')
            indices[:, ix] = [ixp - 1, ixp]
    else:
        ixp = numpy.floor((x - xp[0]) / (xp[1] - xp[0]))
        indices[0, :] = ixp
        indices[1, :] = ixp + 1

    if period is not None:
        indices = indices % nxp
    else:
        indices = numpy.clip(indices, 0, nxp - 1)
    return indices



from roguewave.wavespectra.spectrum import (
    FrequencyDirectionSpectrum,
    FrequencySpectrum,
    NAME_D,
    NAME_F,
    SPECTRAL_DIMS,
)
from roguewave.tools.math import wrapped_difference
from typing import Sequence, TypeVar, Union, Literal
from xarray import concat, Dataset, DataArray
from numpy import diff, empty
from numpy.typing import NDArray
from numba import njit

_T = TypeVar("_T", FrequencySpectrum, FrequencyDirectionSpectrum)
spec_dims = Literal["frequency", "direction"]


def concatenate_spectra(spectra: Sequence[_T], dim=None, keys=None, **kwargs) -> _T:
    """
    Concatenate along the given dimension. If the dimension does not exist a new dimension will be created. Under the
    hood this calls the concat function of xarray. Named arguments to that function can be applied here as well.

    If dim is set to None - we first flatten the spectral objects - and then join along the flattened dimension.

    :param spectra: A sequence of Frequency Spectra/Frequency Direction Spectra
    :param dim: the dimension to concatenate along
    :return: New combined spectral object.
    """

    # Concatenate the dataset in the spectral objects using the xarray concatenate function
    dataset = Dataset()

    if dim is not None:
        # Why are we doing this?
        dataset[dim] = [x.dataset[dim].values for x in spectra]
    else:
        dim = "linear_index"
        spectra = [x.flatten(dim) for x in spectra]

    for variable_name in spectra[0]:
        if variable_name == dim:
            continue

        dataset[variable_name] = concat(
            [x.dataset[variable_name] for x in spectra], dim=dim, **kwargs
        )

    # Get the class of the input spectra.
    cls = type(spectra[0])

    # Return a class instance.
    return cls(dataset)


def integrate_spectral_data(
    dataset: DataArray, dims: Union[spec_dims, Sequence[spec_dims]]
):
    if isinstance(dims, str):
        dims = [dims]

    for dim in dims:
        if dim not in SPECTRAL_DIMS:
            raise ValueError(
                f"Dimension {dim} is not a spectral dimension, options are {NAME_F} or {NAME_D}"
            )

        if dim not in dataset.coords:
            raise ValueError(
                f"Dataset has no {dim} dimension, dimensions are: {list(dataset.dims)}"
            )

    out = dataset
    if NAME_F in dims:
        out = out.fillna(0).integrate(coord=NAME_F)

    if NAME_D in dims:
        difference = DataArray(
            data=wrapped_difference(
                diff(dataset.direction.values, append=dataset.direction[0]), period=360
            ),
            coords={NAME_D: dataset.direction.values},
            dims=[NAME_D],
        )
        out = (out * difference).sum(dim=NAME_D)

    return out


@njit(cache=True, fastmath=True)
def numba_integrate_spectral_data(data: NDArray, grid):

    frequency_step = grid["frequency_step"]
    direction_step = grid["direction_step"]
    number_of_frequencies = data.shape[-2]
    number_of_directions = data.shape[-1]

    return_value = 0.0
    for frequency_index in range(number_of_frequencies):
        for direction_index in range(number_of_directions):
            return_value += (
                data[frequency_index, direction_index]
                * frequency_step[frequency_index]
                * direction_step[direction_index]
            )
    return return_value


@njit(cache=True, fastmath=True)
def numba_directionally_integrate_spectral_data(data: NDArray, grid):

    direction_step = grid["direction_step"]
    number_of_frequencies = data.shape[-2]
    number_of_directions = data.shape[-1]

    return_value = empty(number_of_frequencies, dtype=data.dtype)
    for frequency_index in range(number_of_frequencies):
        return_value[frequency_index] = 0.0
        for direction_index in range(number_of_directions):
            return_value[frequency_index] += (
                data[frequency_index, direction_index] * direction_step[direction_index]
            )
    return return_value



from typing import TypedDict
from roguewave.wavephysics.balance import Dissipation
from numpy.typing import NDArray
from numpy import pi, abs, cos, sin, sqrt, empty, max, exp
from numba import njit
from roguewave.wavetheory import (
    inverse_intrinsic_dispersion_relation,
    intrinsic_group_velocity,
)
from roguewave.wavephysics.fluidproperties import (
    GRAVITATIONAL_ACCELERATION
)
from roguewave.wavespectra.operations import numba_directionally_integrate_spectral_data


class RomeroWaveBreakingParameters(TypedDict):
    saturation_breaking_constant: float
    saturation_threshold: float
    saturation_integrated_threshold: float
    breaking_probability_constant: float
    gravitational_acceleration: float


class RomeroWaveBreaking(Dissipation):
    name = "st4 dissipation"

    def __init__(self, parameters: RomeroWaveBreakingParameters = None):
        super(RomeroWaveBreaking, self).__init__(parameters)
        self._dissipation_function = romero_dissipation_breaking

    @staticmethod
    def default_parameters() -> RomeroWaveBreakingParameters:
        return RomeroWaveBreakingParameters(
            saturation_breaking_constant=2.5,
            saturation_threshold=0.005,
            saturation_integrated_threshold=0.0011,
            breaking_probability_constant=3.5e-5,
            gravitational_acceleration = GRAVITATIONAL_ACCELERATION
        )


@njit(cache=True)
def romero_saturation(
    variance_density: NDArray,
    group_velocity: NDArray,
    wavenumber: NDArray,
    number_of_frequencies: int,
    number_of_directions: int,
):
    directional_saturation_spec = empty(
        (number_of_frequencies, number_of_directions), dtype="float64"
    )

    for frequency_index in range(number_of_frequencies):
        directional_saturation_spec[frequency_index, :] = (
            variance_density[frequency_index, :]
            * group_velocity[frequency_index]
            * wavenumber[frequency_index] ** 3
            / 2
            / pi
        )

    return directional_saturation_spec


@njit(cache=True)
def romero_dissipation_breaking(
    variance_density: NDArray, depth: NDArray, spectral_grid, parameters
):
    number_of_directions = variance_density.shape[1]
    number_of_frequencies = variance_density.shape[0]

    radian_frequency = spectral_grid["radian_frequency"]
    saturation_threshold = parameters["saturation_threshold"]
    saturation_breaking_constant = parameters["saturation_breaking_constant"]
    saturation_integrated_threshold = parameters["saturation_integrated_threshold"]
    breaking_probability_constant = parameters['breaking_probability_constant']
    gravitational_acceleration = parameters['gravitational_acceleration']

    wavenumber = inverse_intrinsic_dispersion_relation(radian_frequency, depth)
    wave_speed = radian_frequency / wavenumber
    group_velocity = intrinsic_group_velocity(wavenumber, depth)

    directionaL_saturation = romero_saturation(
        variance_density=variance_density,
        group_velocity=group_velocity,
        wavenumber=wavenumber,
        number_of_frequencies=number_of_frequencies,
        number_of_directions=number_of_directions,
    )
    saturation = numba_directionally_integrate_spectral_data(directionaL_saturation,spectral_grid)
    dissipation_rate_per_unit_breaking_crest_length = empty((number_of_frequencies))

    for frequency_index in range(number_of_frequencies):
        delta = sqrt(saturation[frequency_index]) - sqrt(saturation_integrated_threshold)
        if delta > 0.0:
            dissipation_rate_per_unit_breaking_crest_length[frequency_index] = (
                saturation_breaking_constant
                * delta**2.5
                / gravitational_acceleration**2
            )
        else:
            dissipation_rate_per_unit_breaking_crest_length[frequency_index] = 0.0

    probability = breaking_probability(
        directionaL_saturation,
        wavenumber,
        saturation_threshold,
        breaking_probability_constant,
        number_of_frequencies,
        number_of_directions
    )

    saturation_breaking = empty((number_of_frequencies, number_of_directions))

    jacobian = 2*pi/group_velocity
    for frequency_index in range(number_of_frequencies):
        for direction_index in range(number_of_directions):
            saturation_breaking[frequency_index,direction_index] = -(
                jacobian[frequency_index]
                * dissipation_rate_per_unit_breaking_crest_length[frequency_index]
                * probability[frequency_index,direction_index]
                * wave_speed[frequency_index]**5
                / gravitational_acceleration**2
            )

    return saturation_breaking

@njit(cache=True)
def breaking_probability(
    directional_saturation,
    wavenumber,
    saturation_threshold,
    breaking_probability_constant,
    number_of_frequencies,
    number_of_directions,
):
    probability = empty(
        (number_of_frequencies, number_of_directions), dtype="float64"
    )

    for frequency_index in range(number_of_frequencies):
        for direction_index in range(number_of_directions):
            if wavenumber[frequency_index] > 0.0:
                probability[frequency_index,direction_index] = (
                    breaking_probability_constant
                    * exp(
                        -saturation_threshold
                        / directional_saturation[frequency_index,direction_index]
                    )
                    / wavenumber[frequency_index]
                )
            else:
                probability[frequency_index] = 0.0

    return probability


from .spotterapi import get_bulk_wave_data, get_data, get_spectrum, get_spotter_data

from .search_endpoint import search_circle, search_rectangle

from .helper_functions import get_spotter_ids, _get_sofar_api



import logging
import sys

# Logging setup
logging.getLogger(__package__).addHandler(logging.NullHandler())
logger = logging.getLogger(__package__)
logger.setLevel(logging.WARNING)
logger.addHandler(logging.NullHandler())

def set_log_to_file(filename, level=logging.INFO):
    logger.addHandler(logging.FileHandler(filename))
    logger.setLevel(level)

def set_level(level):
    if isinstance(level, int):
        logger.setLevel(level)

    elif isinstance(level, str):
        if level == 'debug':
            logger.setLevel(logging.DEBUG)
        elif level == 'info':
            logger.setLevel(logging.INFO)
        elif level == 'warning':
            logger.setLevel(logging.WARNING)
        else:
            raise ValueError(f'unknown logging level {level}')
    else:
        raise ValueError(f'unknown logging level {level}')

def set_log_to_console(level=logging.INFO):
    logger.addHandler(logging.StreamHandler(stream=sys.stdout))
    set_level(level)


from datetime import datetime, timedelta
from boto3 import resource as boto3_resource
from functools import cache
from json import load
from os.path import expanduser
from roguewave.modeldata.timebase import ModelTimeConfiguration, \
    TimeSlice
from typing import List, Dict, Tuple, Literal, Mapping


# Model private variables
# =============================================================================

# Load the model configuration file.
try:
    with open(expanduser('~/model_configuration.json'), 'rb') as file:
        MODELS = load(file)
except FileNotFoundError as e:
    MODELS = {}


# Classes
# =============================================================================
model_type = Literal['analysis','forecast']

class _RemoteResourceSpecification:
    """
    Class containing the key template and other necessary information to
    reconstruct uri's on the remote resource
    """

    def __init__(self,
                 name: str,
                 uri_path_template: str,
                 filetype: str = 'netcdf',
                 scheme: str = 's3',
                 model_type: model_type = 'analysis',
                 mapping_variable_name_model_to_sofar: Dict[str,str] = None,
                 single_variable_per_file: bool = True,
                 model_time_configuration: dict = None):

        """
        :param name: Model name
        :param bucket: bucket
        :param uri_path_template: template of the AWS key
        :param filetype: filetype of the data
        :param model_time_configuration: Description on the timebase of the
        model (how often is the mode run, at what interval is output etc.).
        """
        self.name = name
        self.scheme = scheme
        self.model_type = model_type
        if not isinstance(uri_path_template,Mapping):
            self._uri_path_template = {"default":uri_path_template}
        else:
            self._uri_path_template = uri_path_template
        self.filetype = filetype
        self.mapping_variable_name_model_to_sofar = \
            mapping_variable_name_model_to_sofar
        self.single_variable_per_file = single_variable_per_file

        if model_time_configuration is None:
            self.model_time_configuration = ModelTimeConfiguration()
        else:
            self.model_time_configuration = ModelTimeConfiguration(
                **model_time_configuration
            )

    def uri_path_template(self, variable):
        if variable not in self._uri_path_template:
            return self._uri_path_template['default']
        else:
            return self._uri_path_template[variable]


    def to_sofar_variable_name(self,variable_name: str) -> str:
        """
        map model variable name to sofar variable name
        :param variable_name: model or sofar variable name
        :return: equivalent sofar_variable_name
        """
        if self.mapping_variable_name_model_to_sofar is not None:
            if variable_name in self.mapping_variable_name_model_to_sofar:
                return self.mapping_variable_name_model_to_sofar[
                    variable_name]
            else:
                # if not in the forward mapping, this may already be a sofar
                # variable name, see if this maps back to a model name, if so
                # this is a variable available in this model.
                try:
                    _ = self.to_model_variable_name(variable_name)
                    return variable_name

                except KeyError as e:
                    raise KeyError( f" Modem {self.name} does not have "
                                    f" output that corresponds to a variable "
                                    f" with name {variable_name}" )
        else:
            # if no variable definitions are available, we assume that this is
            # a sofar model, and the mapping is 1 to 1 i.e.
            #       model_variable_name = sofar_variable_name
            return variable_name

    def to_model_variable_name(self,variable_name: str) -> str:
        """
        map sofar variable name to model variable name
        :param variable_name: sofar or model variable name
        :return: equivalent model_variable_name
        """
        if self.mapping_variable_name_model_to_sofar is not None:
            for key,item in self.mapping_variable_name_model_to_sofar.items():
                if item==variable_name:
                    return key
                elif key==variable_name:
                    return key
            else:
                raise KeyError(f'no mapping for variable {variable_name} for model '
                               f'{self.name}')
        else:
            # if no variable definitions are available, we assume that this is
            # a sofar model, and the mapping is 1 to 1 i.e.
            #       model_variable_name = sofar_variable_name
            return variable_name


# Main Functions
# =============================================================================


def available_models() -> List[str]:
    """
    List all available models
    :return: Available model names as list
    """
    return list(MODELS.keys())


def model_timebase( time_slice:TimeSlice, model:str
                    ) -> List[Tuple[datetime, timedelta]]:
    """
    Get all the valid times for a given time slice for the requested model.
    Valid times are returned as a List of pairs: (init_time, forecast_hour)

    :param time_slice: time slice
    :param model: name of the model to retrieve
    :return: List of (init_time, forecast_hour) so that valid_time =
        init_time + forecast_hour
    """
    resource = _get_resource_specification(model)
    return time_slice.time_base(resource.model_time_configuration)


def model_valid_time( time_slice:TimeSlice, model:str) -> List[datetime]:
    """
    Get all the valid times for a given time slice for the requested model.
    Valid times are returned as a List of pairs: (init_time, forecast_hour)

    :param time_slice: time slice
    :param model: name of the model to retrieve
    :return: List of (init_time, forecast_hour) so that valid_time =
        init_time + forecast_hour
    """
    timebase = model_timebase(time_slice,model)
    return [ x[0] + x[1] for x in timebase ]

# -----------------------------------------------------------------------------


@cache
def list_available_variables(model_name, init_time: datetime = None):
    """
    List available variables from a given model. This function _only_ works
    if the variable names are encoded in the aws key. For e.g. ECMWF this
    currently does _not_ work.

    Basically, we look on AWS for the pattern given in model.key_template and
    find all keys conforming to the template with different values for the
    variable name. May be very slow if the prefix (everything before the
    variable name in the template) does not sufficiently restrict the search
    space.

    Note response is cached so that multiple calls are fast.

    :param model_name: the name of the model
    :param init_time: init_time, an init_time for which results from the model
    are available. Defaults to 2022-06-01; this may not work for all models.

    :return: list of variables for the model.
    """

    # This is ugly, but avoids a circular import. ToDo: redo this function.
    from roguewave.modeldata.keygeneration import _generate_aws_key

    init_time = datetime(2022, 6, 1) if init_time is None else init_time
    # get model aws key template description
    model = _get_resource_specification(model_name)

    if model.mapping_variable_name_model_to_sofar is not None:
        return list(model.mapping_variable_name_model_to_sofar.keys())

    # open s3 resource
    s3 = boto3_resource('s3')


    if 'variable' not in model.uri_path_template('default'):
        raise ValueError('Listing variables for a particular model only works'
                         'if the variable name is part of the aws keys.')

    # replace "variable" with "####" as a seperator between everything before
    # and after the variable name in the aws key.
    key = _generate_aws_key(
        '####', init_time, forecast_hour=timedelta(hours=0), model=model)

    # Split in prefix and suffix, and remove the bucket name from the prefix
    prefix, suffix = key.split('####')
    prefix = prefix.replace('s3://','')
    bucket,prefix = prefix.split('/',1)


    # loop over all objects conforming to the prefix
    variables = []
    my_bucket = s3.Bucket(bucket)
    for obj in my_bucket.objects.filter(Prefix=prefix):
        if suffix in obj.key:
            # If the suffix is part of the key, add the part in between the
            # prefix and the suffix of the aws key as a variable name to the
            # output list.
            variables.append(
                _find_between(obj.key, prefix, suffix)
            )

    return variables


# Helper functions
# =============================================================================


def _get_resource_specification(name) -> _RemoteResourceSpecification:
    """
    Return the object representing key layout on aws
    :param name:
    :return: Model
    """
    return _RemoteResourceSpecification(name=name, **MODELS[name])


# -----------------------------------------------------------------------------


def _find_between(s: str, first: str, last: str) -> str:
    """
     Simple function that returns the string between the "first" and "last"
     token. E.g. if first='{' and last=']' with string 'asdd{THIS]asdsd'
     this returns 'THIS'.
    :param s: string
    :param first: token signifying start
    :param last:  token signifying end
    :return:  string
    """

    # lets avoid regex for now :-)
    try:
        start = s.index(first) + len(first)
        end = s.index(last, start)
        return s[start:end]
    except ValueError:
        return ""



from roguewave.interpolate.points import interpolate_points_nd
from roguewave.interpolate.nd_interp import NdInterpolator
from roguewave.interpolate.geometry import Cluster
import numpy

def interpolate_cluster(
    latitude: numpy.ndarray,
    longitude: numpy.ndarray,
    cluster:Cluster,
    get_data,
    period_data=None,
    discont=None
    ):

    coordinates = [('latitude', latitude), ('longitude', longitude)]
    _points = {'latitude': cluster.latitude,
               'longitude': cluster.longitude}

    interpolator = NdInterpolator(
        get_data=get_data,
        data_coordinates=coordinates,
        data_shape=( len(latitude),len(longitude) ),
        interp_coord_names=list(_points.keys()),
        interp_index_coord_name='latitude',
        data_periodic_coordinates={'longitude':360},
        data_period=period_data,
        data_discont=discont
    )

    interpolated_points = interpolator.interpolate(_points)

    output = {}
    for index,name in enumerate(cluster.points.keys()):
        output[name] = interpolated_points[index]

    return output


import typing
from collections import OrderedDict
from dataclasses import dataclass
from datetime import datetime, timezone, timedelta
import numpy
from .environment import http_get
from .utils import bucket_key

TEMPORARY_MAPPING = {
    "Wave_directional_width": "meanDirectionalSpread",
    "Mean_period_of_wind_waves": "peakPeriodWindWaves",
    "Mean_wave_period_of_first_swell_partition": "peakPeriodFirstSwell",
    "Mean_wave_period_of_second_swell_partition": "peakPeriodSecondSwell",
    "Mean_wave_period_of_third_swell_partition": "peakPeriodThirdSwell",
}

NAME_GFS = "GFS"
NAME_WW3 = "SofarOperationalWaveModel"
NAME_HYCOM = "HYCOM"

DEFAULT_MODEL_PER_CATAGORY = {
    "surfaceIce": NAME_GFS,
    "circulation": NAME_HYCOM,
    "surfaceWaves": NAME_WW3,
    "atmospheric": NAME_GFS,
}


@dataclass
class VariableInformation:
    sofar_name: str
    model_id: str
    catagory: str

    @property
    def model_cycle_period_hours(self):
        if self.model_id == "HYCOM":
            return 24
        elif self.model_id == "SofarOperationalWaveModel":
            return 6
        elif self.model_id == "GFS":
            return 6
        elif self.model_id == "GDAS":
            return 6
        else:
            Exception(f"unknown model {self.model_id}")

    @property
    def max_lead_time(self) -> timedelta:
        if self.model_id == "HYCOM":
            return timedelta(days=5)
        elif self.model_id == "SofarOperationalWaveModel":
            return timedelta(days=16)
        elif self.model_id == "GFS":
            return timedelta(days=16)
        elif self.model_id == "GDAS":
            return timedelta(days=1)
        else:
            Exception(f"unknown model {self.model_id}")

    @property
    def model_cycle_period_seconds(self):
        return self.model_cycle_period_hours * 3600

    def model_cycle_start_hours_zulu(self):
        if self.model_id == "HYCOM":
            return 12
        elif self.model_id == "SofarOperationalWaveModel":
            return 0
        elif self.model_id == "GFS":
            return 0
        elif self.model_id == "GDAS":
            return 0
        else:
            Exception(f"unknown model {self.model_id}")

    def nearest_init_time(self, valid_time: datetime) -> datetime:
        valid_time = valid_time.astimezone(timezone.utc)
        time = valid_time.time()
        init_time = (
            int(
                int((time - self.model_cycle_start_hours_zulu() * 3600) / self.model_cycle_period_seconds)
                * self.model_cycle_period_seconds
            )
            + self.model_cycle_start_hours_zulu() * 3600
        )
        return datetime.fromtimestamp(init_time, tz=timezone.utc)

    def nearest_valid_time(self, valid_time: datetime) -> datetime:
        timestamp = valid_time.astimezone(tz=timezone.utc).timestamp()
        return datetime.fromtimestamp(
            (timestamp // (self.output_interval_hours() * 3600)) * self.output_interval_hours() * 3600
        )

    def valid_time_to_forecast_hour(self, time: datetime, lead_time: timedelta = None) -> typing.Tuple[datetime, int]:

        timestamp = time.astimezone(timezone.utc).timestamp()
        init_time = self.nearest_init_time(time)

        def round(i):
            # Avoid bankers rounding
            f = numpy.floor(i)
            return f if i - f < 0.5 else f + 1

        if lead_time:

            number_of_cycles = round(lead_time.total_seconds() / self.model_cycle_period_seconds)

            init_time = init_time - timedelta(seconds=number_of_cycles * self.model_cycle_period_seconds)

            if (
                time - init_time
            ).total_seconds() % self.model_cycle_period_seconds >= 0.5 * self.model_cycle_period_seconds:
                init_time = init_time + timedelta(seconds=self.model_cycle_period_seconds)

        hour = (timestamp - init_time.timestamp()) // 3600
        return init_time, int(hour)

    def output_interval_hours(self):
        if self.model_id == "HYCOM":
            return 3
        elif self.model_id == "SofarOperationalWaveModel":
            return 1
        elif self.model_id == "GFS":
            return 1
        elif self.model_id == "GDAS":
            return 1
        else:
            Exception(f"unknown model {self.model_id}")

    def variable_resolution(self):
        if self.model_id == "HYCOM":
            return "0p08"
        elif self.model_id == "SofarOperationalWaveModel":
            return "0p5"
        elif self.model_id == "GFS":
            return "0p25"
        elif self.model_id == "GDAS":
            return "0p25"
        else:
            Exception(f"unknown model {self.model_id}")


class raster:
    def __init__(self, data_dict):
        self.available = data_dict["available"]
        if self.available:
            self.dataRange = data_dict["dataRange"]
        else:
            self.dataRange = None


class OutputTimes:
    def __init__(self, data_dict):
        self.frequencyHours = data_dict["frequencyHours"]
        self.offset = data_dict["offset"]


class Stub:
    def __init__(self, data_dict):
        self.available = data_dict["available"]


class StormglassVariable:
    def __init__(self, data_dict):
        self.variableID = data_dict["variableID"]
        self.variableName = data_dict["variableName"]
        self.defaultPhysicalUnit = data_dict["defaultPhysicalUnit"]
        self.dataCategory = data_dict["dataCategory"]
        self.json = Stub(data_dict["json"])
        self.netcdf = Stub(data_dict["netcdf"])


class StormglassModelInfo:
    def __init__(self, data_dict):
        self.outputTimes = OutputTimes(data_dict["outputTimes"])
        self.modelID = data_dict["modelID"]
        self.description = data_dict["description"]
        self.variables = []
        for variable in data_dict["variables"]:
            self.variables.append(StormglassVariable(variable))

    def __str__(self):
        data = [x.variableName for x in self.variables]
        return "\n".join(data)


class InitTimes:
    def __init__(self, data_dict):
        self.variableID = data_dict["variableID"]
        self.variableName = data_dict["variableName"]
        self.initTime = data_dict["initTime"]
        self.availableAt = data_dict["availableAt"]

    def initTimeEpoch(self):
        formatting = "%Y-%m-%dT%H:%M:%S.%f%z"
        return int(datetime.timestamp(datetime.strptime(self.initTime, formatting)))


class StormglassInitTime:
    def __init__(self, data_dict):
        self.initTimes = [InitTimes(initTime) for initTime in data_dict["initTimes"]]
        self.dataFormat = data_dict["dataFormat"]
        self.requestTime = data_dict["requestTime"]
        self.modelID = data_dict["modelID"]


def get_gfs_info() -> StormglassModelInfo:
    return get_info(NAME_GFS)


def get_ww3_info() -> StormglassModelInfo:
    return get_info(NAME_WW3)


def get_hycom_info() -> StormglassModelInfo:
    return get_info(NAME_HYCOM)


def get_output_times(model_id, start: datetime, end: datetime, environment=None):
    start = datetime_to_iso_time_string(start)
    end = datetime_to_iso_time_string(end)

    result = http_get(
        f"models/{model_id}/outputTimes?" f"start={start}&end={end}",
        environment,
    ).json()
    return [to_datetime(output_time) for output_time in result["outputTimes"]]


def get_models():
    return http_get("models?dataFormat=netcdf").json()


def get_info(model) -> StormglassModelInfo:
    response = http_get(f"models/{model}?dataFormat=netcdf").json()
    return StormglassModelInfo(response)


def get_init_time_stormglass(
    model_id,
    variableIDs,
    timestamp: datetime,
    data_environment="production",
    cache=OrderedDict(),
    environment=None,
) -> StormglassInitTime:
    if not isinstance(variableIDs, str):
        variableIDs = ",".join(variableIDs)

    string_timestamp = timestamp.strftime("%Y-%m-%dT%H:%M:%SZ")
    hash = model_id + timestamp.strftime("%Y-%m-%dT%H")

    # Use a simple ordered dict as a fifo cache to cache queries.
    if hash in cache:
        return cache[hash]

    endpoint = (
        f"models/"
        f"{model_id}/init-times?"
        f"variableIDs={variableIDs}&"
        f"timestamp={string_timestamp}&"
        f"dataFormat=netcdf&"
        f"environment={data_environment}"
    )

    result = http_get(endpoint, environment).json()

    if len(cache) > 100:
        cache.pop(list(cache.keys())[0])

    cache[hash] = StormglassInitTime(result)
    return StormglassInitTime(result)


def get_init_time(
    model_id,
    variableIDs,
    valid_time: datetime = None,
    environment="production",
):
    """
    Get the closest model init time for a given valid_time. If no valid_time is
    given use _now_ as the valid time.

    :param model_id:
    :param variableIDs:
    :param valid_time:
    :param environment:
    :return:
    """
    if valid_time is None:
        valid_time = datetime.now(tz=timezone.utc)

    init_times_sg = get_init_time_stormglass(model_id, variableIDs, valid_time, environment)
    init_times_str = [x.initTime for x in init_times_sg.initTimes]

    return [to_datetime(t) for t in init_times_str]


def download_netcdf(
    file,
    variable: VariableInformation,
    init_time: datetime,
    forecastHour="full_forecast",
    domain="global",
    legacy=False,
    environment="production",
    data_environment="production",
    experiment=None,
):
    resolution = variable.variable_resolution()
    variable_id = variable.sofar_name
    model_id = variable.model_id

    if experiment or data_environment != "production":
        # for "experiments" we construct the key/bucket ourselves
        bucket, key = bucket_key(
            variable_id,
            model_id,
            init_time,
            forecastHour,
            data_environment,
            resolution,
            domain,
            experiment,
        )
        _ = download_file_from_bucket(key, file, bucket)
    else:
        if forecastHour is None:
            hour = "full_forecast"
        else:
            if forecastHour == "full_forecast":
                hour = "full_forecast"
            else:
                hour = "f%03i" % (forecastHour)

        init_time_string = datetime_to_iso_time_string(init_time)

        endpoint = (
            f"models/{model_id}/"
            f"forecast/netcdf/{variable_id}/"
            f"{init_time_string}/"
            f"{domain}?forecastHour={hour}&resolution={resolution}"
        )

        response = http_get(endpoint, environment=environment)
        with open(file, "wb") as f:
            f.write(response.content)

    if legacy == True:
        convert_to_legacy_netcdf(file, file)

    return file


def get_sofar_name(variable):
    endpoint = f"terms/variable?term={variable}"
    return http_get(endpoint).json()["sofarName"]


def get_variable_information(variable, model_id=None, environment=None) -> VariableInformation:
    if variable in TEMPORARY_MAPPING:
        # A Hack to resolve some missing mappings in the api.
        variable = TEMPORARY_MAPPING[variable]

    endpoint = f"terms/variable?term={variable}"
    data = http_get(endpoint, environment=environment).json()

    if model_id is None:
        model_id = DEFAULT_MODEL_PER_CATAGORY[data["category"]]
    else:
        for model in data["models"]:
            if model["soferName"] == model_id:
                break
        else:
            raise Exception(f"no model found for {variable} with model id {model_id}")

    return VariableInformation(
        sofar_name=data["sofarName"],
        model_id=model_id,
        catagory=data["category"],
    )



from .spectrum import FrequencyDirectionSpectrum, create_2d_spectrum
from abc import ABC, abstractmethod
from scipy.special import gamma
from datetime import datetime
import numpy
import typing
from typing import Literal

PHILLIPS_CONSTANT = 0.0081
GRAVITATIONAL_CONSTANT = 9.81

FrequencyShapeOptions = Literal["pm"]
DirectionalShapeOptions = Literal["raised_cosine"]


def pierson_moskowitz_frequency(
    frequency, peak_frequency, alpha=PHILLIPS_CONSTANT, g=GRAVITATIONAL_CONSTANT
):
    """
    Pierson Moskowitz variance-density spectrum with frequency in Hz as
    dependant variable. See e.g. Holthuijsen "Waves in Oceanic Water."

    :param frequency: frequency in Hz (scalar or array)
    :param peak_frequency: peak frequency in Hz
    :param alpha: Phillips constant (default 0.0081)
    :param g: gravitational acceleration (default 9.81)
    :return:
    """
    return (
        alpha
        * g**2
        * (2 * numpy.pi) ** -4
        * frequency**-5
        * numpy.exp(-5 / 4 * (peak_frequency / frequency) ** 4)
    )


def pierson_moskowitz_angular_frequency(
    angular_frequency: typing.Union[numpy.ndarray, float],
    peak_angular_frequency: float,
    alpha=PHILLIPS_CONSTANT,
    g=GRAVITATIONAL_CONSTANT,
):
    """
    Pierson Moskowitz variance-density spectrum with angular frequency (rad/s) as
    dependant variable. See e.g. Holthuijsen "Waves in Oceanic Water."

    :param frequency: frequency in rad/s (scalar or array)
    :param peak_frequency: peak frequency in rad/s
    :param alpha: Phillips constant (default 0.0081)
    :param g: gravitational acceleration (default 9.81)
    :return:
    """
    return (
        alpha
        * g**2
        * angular_frequency**-5
        * numpy.exp(-5 / 4 * (peak_angular_frequency / angular_frequency) ** 4)
    )


class FrequencyShape(ABC):
    @abstractmethod
    def values(self, frequency_hertz: numpy.ndarray) -> numpy.ndarray:
        pass


class DirectionalShape(ABC):
    @abstractmethod
    def values(self, direction_degrees: numpy.ndarray) -> numpy.ndarray:
        pass


class RaisedCosine(DirectionalShape):
    def __init__(self, mean_direction_degrees: float = 0, width_degrees: float = 28.64):
        self._power = self.power(width_degrees)
        self._mean_direction_degrees = mean_direction_degrees
        self._normalization = (
            numpy.pi
            / 180
            * gamma(self._power / 2 + 1)
            / (gamma(self._power / 2 + 1 / 2) * numpy.sqrt(numpy.pi))
        )

    @staticmethod
    def power(width_degrees):
        return 4 / ((numpy.pi * width_degrees / 90) ** 2) - 2

    def values(self, direction_degrees: numpy.ndarray) -> numpy.ndarray:
        angle = (direction_degrees - self._mean_direction_degrees + 180) % 360 - 180
        with numpy.errstate(invalid="ignore", divide="ignore"):
            return numpy.where(
                numpy.abs(angle) <= 90,
                self._normalization * numpy.cos(angle * numpy.pi / 180) ** self._power,
                0,
            )


class PiersonMoskowitz(FrequencyShape):
    def __init__(
        self, peak_frequency_hertz, m0: float = 1, g: float = GRAVITATIONAL_CONSTANT
    ):
        self._peak_frequency_hertz = peak_frequency_hertz
        self._g = g
        self._alpha = self.alpha(m0)

    def alpha(self, m0):
        return m0 * 5 * (2 * numpy.pi * self._peak_frequency_hertz) ** 4 / self._g**2

    def values(self, frequency_hertz: numpy.ndarray) -> numpy.ndarray:
        """
        Pierson Moskowitz variance-density spectrum with frequency in Hz as
        dependant variable. See e.g. Holthuijsen "Waves in Oceanic Water."

        :param frequency: frequency in Hz (scalar or array)
        :param peak_frequency: peak frequency in Hz
        :param alpha: Phillips constant (default 0.0081)
        :param g: gravitational acceleration (default 9.81)
        :return:
        """
        values = numpy.zeros(len(frequency_hertz))
        msk = frequency_hertz > 0
        values[msk] = (
            self._alpha
            * self._g**2
            * (2 * numpy.pi) ** -4
            * frequency_hertz[msk] ** -5
            * numpy.exp(
                -5 / 4 * (self._peak_frequency_hertz / frequency_hertz[msk]) ** 4
            )
        )
        return values


def create_frequency_shape(
    shape: FrequencyShapeOptions, peak_frequency_hertz: float, m0: float = 1
) -> FrequencyShape:
    if shape == "pm":
        return PiersonMoskowitz(peak_frequency_hertz=peak_frequency_hertz, m0=m0)
    else:
        raise ValueError(f"Unknown frequency shape: {shape}")


def create_directional_shape(
    shape: DirectionalShapeOptions,
    mean_direction_degrees: float = 0,
    width_degrees: float = 30,
) -> DirectionalShape:
    if shape == "raised_cosine":
        return RaisedCosine(
            mean_direction_degrees=mean_direction_degrees, width_degrees=width_degrees
        )
    else:
        raise ValueError(f"Unknown frequency shape: {shape}")


def create_parametric_spectrum(
    frequency_hertz: numpy.ndarray,
    frequency_shape: FrequencyShapeOptions,
    peak_frequency_hertz: float,
    significant_wave_height: float,
    direction_degrees: numpy.ndarray,
    direction_shape: DirectionalShapeOptions,
    mean_direction_degrees: float,
    width_degrees: float,
    depth=numpy.inf,
    time: datetime = None,
    latitude: float = None,
    longitude: float = None,
) -> FrequencyDirectionSpectrum:

    m0 = (significant_wave_height / 4) ** 2

    D = create_directional_shape(
        shape=direction_shape,
        mean_direction_degrees=mean_direction_degrees,
        width_degrees=width_degrees,
    ).values(direction_degrees)

    E = create_frequency_shape(
        shape=frequency_shape, peak_frequency_hertz=peak_frequency_hertz, m0=m0
    ).values(frequency_hertz)

    return create_2d_spectrum(
        frequency=frequency_hertz,
        direction=direction_degrees,
        variance_density=E[:, None] * D[None, :],
        time=time,
        latitude=latitude,
        longitude=longitude,
        depth=depth,
        dims=("frequency", "direction"),
    )



from typing import TypedDict
from roguewave.wavephysics.balance import Dissipation
from numpy.typing import NDArray
from numpy import pi, abs, cos, sin, sqrt, empty, max
from numba import njit
from roguewave.wavetheory import (
    inverse_intrinsic_dispersion_relation,
    intrinsic_group_velocity,
)


class ST4WaveBreakingParameters(TypedDict):
    saturation_breaking_constant: float
    saturation_breaking_directional_control: float
    saturation_cosine_power: float
    saturation_integration_width_degrees: float
    saturation_threshold: float
    cumulative_breaking_constant: float
    cumulative_breaking_max_relative_frequency: float


class ST4WaveBreaking(Dissipation):
    name = "st4 dissipation"

    def __init__(self, parameters: ST4WaveBreakingParameters = None):
        super(ST4WaveBreaking, self).__init__(parameters)
        self._dissipation_function = st4_dissipation_breaking

    @staticmethod
    def default_parameters() -> ST4WaveBreakingParameters:
        return ST4WaveBreakingParameters(
            saturation_breaking_constant=2.2e-05,
            saturation_breaking_directional_control=0,
            saturation_cosine_power=2,
            saturation_integration_width_degrees=80,
            saturation_threshold=0.0009,
            cumulative_breaking_constant=0.4,
            cumulative_breaking_max_relative_frequency=0.5,
        )


@njit(cache=True)
def st4_band_integrated_saturation(
    variance_density: NDArray,
    group_velocity: NDArray,
    wavenumber: NDArray,
    radian_direction: NDArray,
    direction_step: NDArray,
    number_of_frequencies: int,
    number_of_directions: int,
    integration_width_degrees: int,
    cosine_power=2,
):
    directional_saturation_spec = empty(
        (number_of_frequencies, number_of_directions), dtype="float64"
    )
    band_integrated_saturation = empty(
        (number_of_frequencies, number_of_directions), dtype="float64"
    )
    integration_width_radians = integration_width_degrees * pi / 180

    for frequency_index in range(number_of_frequencies):
        directional_saturation_spec[frequency_index, :] = (
            variance_density[frequency_index, :]
            * group_velocity[frequency_index]
            * wavenumber[frequency_index] ** 3
            / 2
            / pi
        )

    for frequency_index in range(number_of_frequencies):
        for direction_index in range(number_of_directions):
            integrant = 0.0
            for dummy_direction_index in range(number_of_directions):
                mutual_angle = (
                    radian_direction[dummy_direction_index]
                    - radian_direction[direction_index]
                    + pi
                ) % (2 * pi) - pi
                if abs(mutual_angle) > integration_width_radians:
                    continue

                integrant += (
                    directional_saturation_spec[frequency_index, dummy_direction_index]
                    * cos(mutual_angle) ** cosine_power
                    * direction_step[dummy_direction_index]
                )
            band_integrated_saturation[frequency_index, direction_index] = integrant
    return band_integrated_saturation


@njit(cache=True)
def st4_cumulative_breaking(
    variance_density: NDArray,
    saturation: NDArray,
    radian_frequency: NDArray,
    group_velocity: NDArray,
    wave_speed: NDArray,
    radian_direction: NDArray,
    direction_step: NDArray,
    frequency_step: NDArray,
    saturation_threshold: float,
    cumulative_breaking_constant: float,
    cumulative_breaking_max_relative_frequency: float,
    number_of_frequencies: int,
    number_of_directions: int,
):
    """

    :param saturation:
    :param radian_frequency:
    :param group_velocity:
    :param wave_speed:
    :param radian_direction:
    :param direction_step:
    :param frequency_step:
    :param saturation_threshold:
    :param cumulative_breaking_max_relative_frequency:
    :param number_of_frequencies:
    :param number_of_directions:
    :return:
    """

    jacobian_radian_to_degrees_direction = pi / 180
    jacobian_radian_to_hertz_frequency = 2 * pi
    jacobian_wavenumber_to_radian_frequency = 1 / group_velocity

    jacobian = (
        jacobian_radian_to_hertz_frequency
        * jacobian_wavenumber_to_radian_frequency
        * jacobian_radian_to_degrees_direction
    )

    cosines = cos(radian_direction)
    sines = sin(radian_direction)

    threshold = sqrt(saturation) - sqrt(saturation_threshold)

    cumulative_breaking = empty(
        (number_of_frequencies, number_of_directions), dtype="float64"
    )
    if not (cumulative_breaking_constant > 0.0):
        cumulative_breaking[:, :] = 0.0
        return cumulative_breaking

    wave_speed_east = empty(
        (number_of_frequencies, number_of_directions), dtype="float64"
    )
    wave_speed_north = empty(
        (number_of_frequencies, number_of_directions), dtype="float64"
    )

    for frequency_index in range(number_of_frequencies):
        for direction_index in range(number_of_directions):
            wave_speed_east[frequency_index, direction_index] = (
                cosines[direction_index] * wave_speed[frequency_index]
            )
            wave_speed_north[frequency_index, direction_index] = (
                sines[direction_index] * wave_speed[frequency_index]
            )

    for frequency_index in range(number_of_frequencies):
        for direction_index in range(number_of_directions):
            # Initialize
            strength = 0.0

            for dummy_frequency_index in range(0, number_of_frequencies):
                #
                if (
                    radian_frequency[dummy_frequency_index]
                    > radian_frequency[frequency_index]
                    * cumulative_breaking_max_relative_frequency
                ):
                    break

                for dummy_direction_index in range(0, number_of_directions):
                    if threshold[dummy_frequency_index, dummy_direction_index] <= 0:
                        continue

                    integrant = (
                        threshold[dummy_frequency_index, dummy_direction_index] ** 2
                        * jacobian[dummy_frequency_index]
                    )

                    # Calculate the celerity magnitude
                    delta_wave_speed_east = (
                        wave_speed_east[dummy_frequency_index, dummy_direction_index]
                        - wave_speed_east[frequency_index, direction_index]
                    )
                    delta_wave_speed_north = (
                        wave_speed_north[dummy_frequency_index, dummy_direction_index]
                        - wave_speed_north[frequency_index, direction_index]
                    )
                    delta_speed_magnitude = sqrt(
                        delta_wave_speed_east**2 + delta_wave_speed_north**2
                    )

                    strength += (
                        frequency_step[dummy_frequency_index]
                        * direction_step[dummy_direction_index]
                        * delta_speed_magnitude
                        * integrant
                    )

            cumulative_breaking[frequency_index, direction_index] = (
                -1.44
                * cumulative_breaking_constant
                * strength
                * variance_density[frequency_index, direction_index]
            )
    return cumulative_breaking


@njit(cache=True)
def st4_dissipation_breaking(
    variance_density: NDArray, depth: NDArray, spectral_grid, parameters
):
    number_of_directions = variance_density.shape[1]
    number_of_frequencies = variance_density.shape[0]

    radian_frequency = spectral_grid["radian_frequency"]
    radian_direction = spectral_grid["radian_direction"]
    frequency_step = spectral_grid["frequency_step"]
    direction_step = spectral_grid["direction_step"]

    saturation_integration_width_degrees = parameters[
        "saturation_integration_width_degrees"
    ]
    saturation_cosine_power = parameters["saturation_cosine_power"]
    saturation_threshold = parameters["saturation_threshold"]
    cumulative_breaking_constant = parameters["cumulative_breaking_constant"]
    cumulative_breaking_max_relative_frequency = parameters[
        "cumulative_breaking_max_relative_frequency"
    ]
    saturation_breaking_constant = parameters["saturation_breaking_constant"]
    saturation_breaking_directional_control = parameters[
        "saturation_breaking_directional_control"
    ]

    wavenumber = inverse_intrinsic_dispersion_relation(radian_frequency, depth)
    wave_speed = radian_frequency / wavenumber
    group_velocity = intrinsic_group_velocity(wavenumber, depth)

    band_integrated_saturation = st4_band_integrated_saturation(
        variance_density=variance_density,
        group_velocity=group_velocity,
        wavenumber=wavenumber,
        radian_direction=radian_direction,
        direction_step=direction_step,
        number_of_frequencies=number_of_frequencies,
        number_of_directions=number_of_directions,
        integration_width_degrees=saturation_integration_width_degrees,
        cosine_power=saturation_cosine_power,
    )

    cumulative_breaking = st4_cumulative_breaking(
        variance_density=variance_density,
        saturation=band_integrated_saturation,
        radian_frequency=radian_frequency,
        group_velocity=group_velocity,
        wave_speed=wave_speed,
        radian_direction=radian_direction,
        direction_step=direction_step,
        frequency_step=frequency_step,
        saturation_threshold=saturation_threshold,
        cumulative_breaking_constant=cumulative_breaking_constant,
        cumulative_breaking_max_relative_frequency=cumulative_breaking_max_relative_frequency,
        number_of_frequencies=number_of_frequencies,
        number_of_directions=number_of_directions,
    )

    saturation_breaking = st4_saturation_breaking(
        variance_density=variance_density,
        band_integrated_saturation=band_integrated_saturation,
        radian_frequency=radian_frequency,
        number_of_frequencies=number_of_frequencies,
        number_of_directions=number_of_directions,
        saturation_breaking_constant=saturation_breaking_constant,
        saturation_breaking_directional_control=saturation_breaking_directional_control,
        saturation_threshold=saturation_threshold,
    )

    return cumulative_breaking + saturation_breaking


@njit(cache=True)
def st4_saturation_breaking(
    variance_density,
    band_integrated_saturation,
    radian_frequency,
    number_of_frequencies,
    number_of_directions,
    saturation_breaking_constant,
    saturation_breaking_directional_control,
    saturation_threshold,
):
    saturation_breaking = empty(
        (number_of_frequencies, number_of_directions), dtype="float64"
    )

    if not (saturation_breaking_constant > 0.0):
        saturation_breaking[:, :] = 0.0
        return saturation_breaking

    for frequency_index in range(number_of_frequencies):
        maximum_band_integrated_saturation = max(
            band_integrated_saturation[frequency_index, :]
        )
        isotropic_saturation_exceedence = (
            (maximum_band_integrated_saturation - saturation_threshold)
            / saturation_threshold
            * saturation_breaking_directional_control
        )
        if isotropic_saturation_exceedence < 0.0:
            isotropic_saturation_exceedence = 0.0

        for direction_index in range(number_of_directions):
            directional_saturation_exceedence = (
                (
                    band_integrated_saturation[frequency_index, direction_index]
                    - saturation_threshold
                )
                / saturation_threshold
                * (1 - saturation_breaking_directional_control)
            )
            if directional_saturation_exceedence < 0.0:
                directional_saturation_exceedence = 0.0

            relative_exceedence_level = (
                isotropic_saturation_exceedence + directional_saturation_exceedence
            )

            saturation_breaking[frequency_index, direction_index] = (
                -saturation_breaking_constant
                * relative_exceedence_level**2
                * radian_frequency[frequency_index]
                * variance_density[frequency_index, direction_index]
            )
    return saturation_breaking



"""
Contents: Simple file caching routines to interact with a file cache.

Copyright (C) 2022
Sofar Ocean Technologies

Authors: Pieter Bart Smit
======================

Functions:
- `filepaths`, given URI's return a filepath to the locally stored
   version
- `exists`, does a cache with a given name exists
- `create_cache`, create a cache with a given name and custom properties.
- `delete_cache`, delete files associated with the cache.
- `delete_default`, delete files associated with the default cache.
- `delete_files`, remove entries from a given cache.
- `_get_cache`, get Cache object corresponding to the name (for internal use
   only)
"""

# Import
# =============================================================================
import os
from typing import List, Tuple, Union, Dict, Iterable, Callable
from .cache_object import TEMPORARY_DIRECTORY, CACHE_SIZE_GB, FileCache
from .remote_resources import RemoteResource

# Constants
# =============================================================================


DEFAULT_CACHE_NAME = "__default__"

# Private Module Variables
# =============================================================================


# This dictionary contains all instantiated FileCache objects as values and
# the object name as key.
_ACTIVE_FILE_CACHES = {}  # type: Dict[str,FileCache]


# Main public function.
# =============================================================================
def set(name, value, cache_name: str = None):
    cache = get_cache(cache_name)
    setattr(cache, name, value)


def filepaths(
    uris: Union[List[str], str],
    cache_name: str = None,
) -> Union[List[str], Tuple[List[str], List[bool]]]:
    """
    Return the full file path to locally stored objects corresponding to the
    given URI

    :param uris: List of uris, or a single uri
    :param cache_name: name of the cache to use. If None, a default cache will
    be initialized automatically (if not initialized) and used.
    :param return_cache_hits: return whether or not the files were already in
        cache or downloaded from the remote source (cache hit or miss).

    :return: List Absolute paths to the locally stored versions corresponding
        to the list of URI's. IF return_cache_hits=True, additionally return
        a list of cache hits as the second entry of the return tuple.
    """
    return get_cache(cache_name)[uris]


def remove_directive_function(directive: str, name: str, cache_name=None):
    _ = get_cache(cache_name).remove_directive_function(directive, name)


def set_directive_function(
    directive: str,
    name: str,
    post_process_function: Union[Callable[[str], None], Callable[[str], bool]] = None,
    cache_name=None,
):
    #
    _ = get_cache(cache_name).set_directive_function(
        directive, name, post_process_function
    )


def exists(cache_name: str):
    """
    Check if the cache name already exists
    :param cache_name: name for the cache to be created. This name is used
            to retrieve files from the cache.
    :return: True if exists, False otherwise
    """
    return cache_name in _ACTIVE_FILE_CACHES


def create_cache(
    cache_name: str,
    cache_path: str = TEMPORARY_DIRECTORY,
    cache_size_GB: Union[int, float] = CACHE_SIZE_GB,
    do_cache_eviction_on_startup: bool = False,
    download_in_parallel=True,
    resources: List[RemoteResource] = None,
) -> None:
    """
    Create a file cache. Created caches *must* have unique names and
    cache_paths.

    :param cache_name: name for the cache to be created. This name is used
            to retrieve files from the cache.
    :param cache_path: path to store cache. If path does not exist it will be
            created.
    :param cache_size_GB:  Maximum size of the cache in GiB. If cache exceeds
            the size, then files with oldest access/modified dates get deleted
            until everthing fits in the cache again. Fractional values (floats)
            are allowed.
    :param do_cache_eviction_on_startup: do_cache_eviction_on_startup: whether
            we ensure the cache size conforms to the given size on startup.
            If set to true, a cache directory that exceeds the maximum size
            will be reduced to max size. Set to False by default in which case
            an error occurs. The latter to prevent eroneously evicting files
            from a cache that was previously created on purpose with a larger
            size that the limit.
    :param download_in_parallel: Download in paralel from resource. Per default 10
            worker threads are created.

    :return:
    """
    cache_path = os.path.abspath(os.path.expanduser(cache_path))

    if cache_name in _ACTIVE_FILE_CACHES:
        raise ValueError(f"Cache with name {cache_name} is already initialized")

    for key, cache in _ACTIVE_FILE_CACHES.items():
        if cache.path == cache_path:
            raise ValueError(
                f"Error when creating cache with name: "
                f'"{cache_name}". \n A cache named: "{key}" '
                f"already uses the path {cache_path} "
                f"for caching.\n "
                f"Multiple caches cannot share the same path."
            )

    _ACTIVE_FILE_CACHES[cache_name] = FileCache(
        cache_path,
        size_GB=cache_size_GB,
        do_cache_eviction_on_startup=do_cache_eviction_on_startup,
        parallel=download_in_parallel,
        resources=resources,
    )
    return


def delete_cache(cache_name):
    """
    Delete all files associated with a cache and remove cache from available
    caches. To note: all files are deleted, but the folder itself is not.

    :param cache_name: Name of the cache to be deleted
    :return:
    """
    if not exists(cache_name):
        raise ValueError(f"Cache with name {cache_name} does not exist")

    cache = _ACTIVE_FILE_CACHES.pop(cache_name)
    cache.purge()


def delete_default():
    """
    Clean up the default cache.

    :return:
    """
    if exists(DEFAULT_CACHE_NAME):
        delete_cache(DEFAULT_CACHE_NAME)


def delete_files(
    uris: Union[str, Iterable[str]], cache_name: str = None, error_if_not_in_cache=True
) -> None:
    """
    Remove given key(s) from the cache
    :param uris: list of keys to remove
    :param cache_name: name of initialized cache.
    :return:
    """
    if not isinstance(uris, Iterable) or isinstance(uris, str):
        uris = [uris]

    cache = get_cache(cache_name)

    for key in uris:
        try:
            cache.remove(key)

        except ValueError as e:
            if error_if_not_in_cache:
                raise e


def get_cache(cache_name: str) -> FileCache:
    """
    Get a valid cache object, error if the name does not exist.
    :param cache_name: Name of the cache
    :return: Cache object
    """

    if cache_name is None:
        cache_name = DEFAULT_CACHE_NAME

    if not exists(cache_name):
        if cache_name == DEFAULT_CACHE_NAME:
            create_cache(cache_name)
        else:
            raise ValueError(f"Cache with name {cache_name} does not exist.")

    return _ACTIVE_FILE_CACHES[cache_name]



"""
Contents: Routines to read raw data from Sofar Spotters

Copyright (C) 2022
Sofar Ocean Technologies

Authors: Pieter Bart Smit
======================

Routines read csv data stored on SD-Cards and return in a convenient form.

Public Functions:

- `read_baro`, read ????_BARO.csv files and return a dataframe.
- `read_baro_raw`, read ????_BARO_RAW.csv files and return a dataframe.
- `read_gmn`, read ????_GMN.csv files and return a dataframe.
- `read_raindb`, read ????_RAINDB.csv files and return a dataframe.
- `read_sst`, read ????_SST.csv files and return a dataframe.
- `read_displacement`, read ????_FLT.csv files that contain displacement data
- `read_gps`, read ????_GPS.csv files that contain raw GPS strings.
- `read_location`, read ????_LOC.csv files that containt the location if the instrument.
- `read_raw_spectra`, read ????_SPC.csv files that contain raw spectral data.
- `read_spectra`, read ????_SPC.csv files and return a spectral object.

"""

from pandas import DataFrame, concat, to_datetime

from roguewave.spotter.parser import (
    read_and_concatenate_spotter_csv,
    get_csv_file_format,
    apply_to_group,
)
from roguewave.timeseries_analysis.filtering import sos_filter
from xarray import Dataset
from roguewave import FrequencySpectrum
from roguewave.tools.time import datetime64_to_timestamp
from datetime import datetime
from numpy import linspace, errstate, sqrt, interp, full_like, inf, cos, sin, pi, nan


# Main Functions
# ---------------------------------


def read_gps(
    path,
    start_date: datetime = None,
    end_date: datetime = None,
    postprocess=True,
    cache_as_netcdf=False,
) -> DataFrame:
    """
    Load raw GPS text files and return a pandas dataframe containing the data. By default the data is postprocessed into
    a more  convinient form (without loss of information) unless raw data is specifically requested

    :param path: Path containing Spotter CSV files
    :param postprocess: whether to postprocess the data. Postprocessing converts heading and velocity magnitude to
                        velocity components, and combines latitude and lituted minutes into a single double latitude
                        (same for longitudes).

    :param start_date: If only a subset of the data is needed we can avoid loading all data, this denotes the start
                       date of the desired interval. If given, only data after the start_date is loaded (if available).
                       NOTE: this requires that LOC files are present.

    :param end_date: If only a subset of the data is needed we can avoid loading all data, this denotes the end
                       date of the desired interval. If given, only data before the end_date is loaded (if available).
                       NOTE: this requires that LOC files are present.

    :return: Pandas Dataframe. If postprocess is false it just contains the raw columns of the GPS file (see file for
             description) If True (default) returns dataframe with columns

             "time": epoch time (UTC, epoch of 1970-1-1, i.e. Unix Epoch).
             "latitude": Latitude in decimal degrees
             "longitude": Longitude in decimal degrees
             "z": raw vertical elevation from GPS in meter
             "u': eastward velocity, m/s
             "v': northward velocity, m/s
             "w": vertical velocity, m/s
             "group id": Identifier that indicates continuous data groups. (data from "same deployment).

    """

    data = read_and_concatenate_spotter_csv(
        path,
        "GPS",
        start_date=start_date,
        end_date=end_date,
        cache_as_netcdf=cache_as_netcdf,
    )
    if not postprocess:
        return data
    u = (
        data["SOG(mm_s)"].values
        / 1000
        * cos((90 - data["COG(deg*1000)"].values / 1000) * pi / 180)
    )
    v = (
        data["SOG(mm_s)"].values
        / 1000
        * sin((90 - data["COG(deg*1000)"].values / 1000) * pi / 180)
    )
    w = data["vert_vel(mm_s)"].values / 1000
    time = to_datetime(data["GPS_Epoch_Time(s)"].values, unit="s")

    latitude = data["lat(deg)"].values + data["lat(min*1e5)"].values / 60e5
    longitude = data["long(deg)"].values + data["long(min*1e5)"].values / 60e5
    z = data["el(m)"].values

    return DataFrame(
        data={
            "time": time,
            "latitude": latitude,
            "longitude": longitude,
            "z": z,
            "u": u,
            "v": v,
            "w": w,
            "group id": data["group id"].values,
        }
    )


def read_displacement(
    path: str,
    start_date: datetime = None,
    end_date: datetime = None,
    postprocess: bool = True,
    cache_as_netcdf=False,
) -> DataFrame:

    """
    Load displacement dataand return a pandas dataframe containing the data.
    By default the data is postprocessed to apply an inverse pass of the
    IIR filter to correct for phase differences.

    :param path: Path containing Spotter CSV files

    :param start_date: If only a subset of the data is needed we can avoid
                       loading all data, this denotes the start date of the
                       desired interval. If given, only data after the
                       start_date is loaded (if available).
                       NOTE: this requires that LOC files are present.

    :param end_date: If only a subset of the data is needed we can avoid
                     loading all data, this denotes the end date of the
                     desired interval. If given, only data before the
                     end_date is loaded (if available).
                     NOTE: this requires that LOC files are present.

    :param postprocess: whether to apply the phase correction

    :return: Pandas Dataframe. Returns dataframe with columns

             "time": epoch time (UTC, epoch of 1970-1-1, i.e. Unix Epoch).
             "x": filteresd displacement data (Eastings)
             "y': filteresd displacement data (Northings)
             "z': vertical displacement from local mean.
             "group id": Identifier that indicates continuous data groups.
                         (data from "same deployment).
    """

    data = read_and_concatenate_spotter_csv(
        path,
        "FLT",
        start_date=start_date,
        end_date=end_date,
        cache_as_netcdf=cache_as_netcdf,
    )
    if not postprocess:
        return data

    # Backward sos pass for phase correction
    def _process(_data: DataFrame):
        _df = DataFrame()
        _df["time"] = _data["time"]
        _df["x"] = sos_filter(_data["outx(mm)"].values / 1000.0, "backward")
        _df["y"] = sos_filter(_data["outy(mm)"].values / 1000.0, "backward")
        _df["z"] = sos_filter(_data["outz(mm)"].values / 1000.0, "backward")
        return _df

    return apply_to_group(_process, data)


def read_baro(
    path: str,
    start_date: datetime = None,
    end_date: datetime = None,
):
    """
    Read filtered barometer files.

    :param path: Path containing Spotter CSV files

    :param start_date: If only a subset of the data is needed we can avoid
                       loading all data, this denotes the start date of the
                       desired interval. If given, only data after the
                       start_date is loaded (if available).
                       NOTE: this requires that LOC files are present.

    :param end_date: If only a subset of the data is needed we can avoid
                     loading all data, this denotes the end date of the
                     desired interval. If given, only data before the
                     end_date is loaded (if available).
                     NOTE: this requires that LOC files are present.
    """
    return read_and_concatenate_spotter_csv(path, "BARO", start_date, end_date)


def read_baro_raw(
    path: str,
    start_date: datetime = None,
    end_date: datetime = None,
):
    """
    Read raw barometer files (no filtering)

    :param path: Path containing Spotter CSV files

    :param start_date: If only a subset of the data is needed we can avoid
                       loading all data, this denotes the start date of the
                       desired interval. If given, only data after the
                       start_date is loaded (if available).
                       NOTE: this requires that LOC files are present.

    :param end_date: If only a subset of the data is needed we can avoid
                     loading all data, this denotes the end date of the
                     desired interval. If given, only data before the
                     end_date is loaded (if available).
                     NOTE: this requires that LOC files are present.
    """
    return read_and_concatenate_spotter_csv(path, "BARO_RAW", start_date, end_date)


def read_sst(
    path: str,
    start_date: datetime = None,
    end_date: datetime = None,
):
    """
    Read SST data.

    :param path: Path containing Spotter CSV files

    :param start_date: If only a subset of the data is needed we can avoid
                       loading all data, this denotes the start date of the
                       desired interval. If given, only data after the
                       start_date is loaded (if available).
                       NOTE: this requires that LOC files are present.

    :param end_date: If only a subset of the data is needed we can avoid
                     loading all data, this denotes the end date of the
                     desired interval. If given, only data before the
                     end_date is loaded (if available).
                     NOTE: this requires that LOC files are present.
    """
    return read_and_concatenate_spotter_csv(path, "sst", start_date, end_date)


def read_raindb(
    path: str,
    start_date: datetime = None,
    end_date: datetime = None,
):
    """
    Read sound volume in decibel.

    :param path: Path containing Spotter CSV files

    :param start_date: If only a subset of the data is needed we can avoid
                       loading all data, this denotes the start date of the
                       desired interval. If given, only data after the
                       start_date is loaded (if available).
                       NOTE: this requires that LOC files are present.

    :param end_date: If only a subset of the data is needed we can avoid
                     loading all data, this denotes the end date of the
                     desired interval. If given, only data before the
                     end_date is loaded (if available).
                     NOTE: this requires that LOC files are present.
    """
    return read_and_concatenate_spotter_csv(path, "RAINDB", start_date, end_date)


def read_gmn(
    path: str,
    start_date: datetime = None,
    end_date: datetime = None,
):
    """
    Read gps metric files- for development purposes only.

    :param path: Path containing Spotter CSV files

    :param start_date: If only a subset of the data is needed we can avoid
                       loading all data, this denotes the start date of the
                       desired interval. If given, only data after the
                       start_date is loaded (if available).
                       NOTE: this requires that LOC files are present.

    :param end_date: If only a subset of the data is needed we can avoid
                     loading all data, this denotes the end date of the
                     desired interval. If given, only data before the
                     end_date is loaded (if available).
                     NOTE: this requires that LOC files are present.
    """
    return read_and_concatenate_spotter_csv(path, "GMN", start_date, end_date)


def read_location(
    path: str,
    start_date: datetime = None,
    end_date: datetime = None,
    postprocess: bool = True,
) -> DataFrame:
    """
    :param path: Path containing Spotter CSV files

    :param start_date: If only a subset of the data is needed we can avoid
                       loading all data, this denotes the start date of the
                       desired interval. If given, only data after the
                       start_date is loaded (if available).
                       NOTE: this requires that LOC files are present.

    :param end_date: If only a subset of the data is needed we can avoid
                     loading all data, this denotes the end date of the
                     desired interval. If given, only data before the
                     end_date is loaded (if available).
                     NOTE: this requires that LOC files are present.

    :param postprocess: whether to apply the phase correction

    :return: Pandas Dataframe. Returns dataframe with columns

             "time": epoch time (UTC, epoch of 1970-1-1, i.e. Unix Epoch).
             "latitude": latitude in decimal degrees
             "longitude': longitude in decimal degrees
             "group id": Identifier that indicates continuous data groups.
                         (data from "same deployment).
    """
    raw_data = read_and_concatenate_spotter_csv(
        path,
        "LOC",
        start_date=start_date,
        end_date=end_date,
    )
    if not postprocess:
        return raw_data

    dataframe = DataFrame()
    dataframe["time"] = raw_data["time"]

    dataframe["latitude"] = (
        raw_data["lat(deg)"].values + raw_data["lat(min*1e5)"].values / 60e5
    )
    dataframe["longitude"] = (
        raw_data["long(deg)"].values + raw_data["long(min*1e5)"].values / 60e5
    )
    dataframe["group id"] = raw_data["group id"]
    return dataframe


def read_spectra(
    path: str,
    start_date: datetime = None,
    end_date: datetime = None,
    depth: float = inf,
    cache_as_netcdf=False,
) -> FrequencySpectrum:
    """
    Read spectral data from csv files. The raw spectral data is transformed into
    a roguewave Spectral 1D spectrum object (which includes all directional moments a1,b1,a2,b2 as well as energy for
    the given time period).

    :param path: Path containing Spotter CSV files
    :param postprocess: Whether to apply the phase correction

    :param start_date: If only a subset of the data is needed we can avoid
                       loading all data, this denotes the start date of the
                       desired interval. If given, only data after the
                       start_date is loaded (if available).
                       NOTE: this requires that LOC files are present.

    :param end_date: If only a subset of the data is needed we can avoid
                     loading all data, this denotes the end date of the
                     desired interval. If given, only data before the
                     end_date is loaded (if available).
                     NOTE: this requires that LOC files are present.

    :param depth: Local water depth,  by default set to inf (deep water).
                  Not required, but is set on the returned spectral object
                  (and factors in transformations thereof, e.g. to get wavenumbers).

    :return: frequency spectra as a FrequencySpectrum object.
    """

    data = read_raw_spectra(
        path, start_date=start_date, end_date=end_date, cache_as_netcdf=cache_as_netcdf
    )
    spectral_file_format = get_csv_file_format("SPC")
    df = 1 / (
        spectral_file_format["nfft"] * spectral_file_format["sampling_interval_gps"]
    )
    frequencies = (
        linspace(
            0,
            spectral_file_format["nfft"] // 2,
            spectral_file_format["nfft"] // 2,
            endpoint=False,
        )
        * df
    )

    spectral_values = data[list(range(0, spectral_file_format["nfft"] // 2))].values
    time = data["time"].values
    time = time[data["kind"] == "Szz_re"]

    Szz = spectral_values[data["kind"] == "Szz_re", :]
    Sxx = spectral_values[data["kind"] == "Sxx_re", :]
    Syy = spectral_values[data["kind"] == "Syy_re", :]
    Cxy = spectral_values[data["kind"] == "Sxy_re", :]
    Qzx = spectral_values[data["kind"] == "Szx_im", :]
    Qzy = spectral_values[data["kind"] == "Szy_im", :]

    with errstate(invalid="ignore", divide="ignore"):
        # Supress divide by 0; silently produce NaN
        a1 = Qzx / sqrt(Szz * (Sxx + Syy))
        a2 = (Sxx - Syy) / (Sxx + Syy)
        b1 = Qzy / sqrt(Szz * (Sxx + Syy))
        b2 = 2.0 * Cxy / (Sxx + Syy)

    try:
        location = read_location(path, postprocess=True)
        latitude = interp(
            datetime64_to_timestamp(time),
            datetime64_to_timestamp(location["time"].values),
            location["latitude"].values,
        )
        longitude = interp(
            datetime64_to_timestamp(time),
            datetime64_to_timestamp(location["time"].values),
            location["longitude"].values,
        )

    except FileNotFoundError:
        # No location files find to get latitude/longitude. Just fill with NaN.
        latitude = full_like(time, nan)
        longitude = full_like(time, nan)

    depth = full_like(time, depth)

    dataset = Dataset(
        data_vars={
            "variance_density": (["time", "frequency"], Szz),
            "a1": (["time", "frequency"], a1),
            "b1": (["time", "frequency"], b1),
            "a2": (["time", "frequency"], a2),
            "b2": (["time", "frequency"], b2),
            "depth": (["time"], depth),
            "latitude": (["time"], latitude),
            "longitude": (["time"], longitude),
        },
        coords={"time": to_datetime(time), "frequency": frequencies},
    )
    return FrequencySpectrum(dataset)


def read_raw_spectra(
    path,
    postprocess=True,
    start_date: datetime = None,
    end_date: datetime = None,
    cache_as_netcdf=False,
) -> DataFrame:
    """
    Read raw spectral files and return a dataframe

    :param path:
    :param postprocess:
    :param start_date:
    :param end_date:
    :return:
    """
    raw_data = read_and_concatenate_spotter_csv(
        path,
        "SPC",
        start_date=start_date,
        end_date=end_date,
        cache_as_netcdf=cache_as_netcdf,
    )
    if not postprocess:
        return raw_data

    # We want to collect all spectra of the same type (Sxx_re or Syy_im) in a single dataframe, with frequencies as
    # rows. First, we list all of the columns associated with a particular variable.
    spec_format = get_csv_file_format("SPC")
    column_names = {}

    # Loop over the columns in the spec format
    for column in spec_format["columns"]:
        # If it is not one of the spectral columns, continue (e.g. time)
        if column["name"][0:3] not in ["Sxx", "Syy", "Szz", "Sxy", "Szx", "Szy"]:
            continue

        # Get the full name of the column (Sxx_re, or Sxy_im) without the frequency enumeration. This is the variable
        # of interest
        name = column["name"][0:6]

        # add full column name to the list of columns belonging to this variable. e.g. we get for
        # {
        #   "Szz_re": ["Szz_re_0", "Szz_re_1", ..., "Szz_re_128"],
        #   ....         ....         ....     ....       ....
        #   "Szy_re": ["Szy_im_0", "Szy_im_1", ..., "Szy_im_128"]
        # }
        if name not in column_names:
            column_names[name] = []

        column_names[name].append(column["name"])

    # initialize
    dataframes = []
    data_types = []

    frequency_resolution = 1 / (
        spec_format["sampling_interval_gps"] * spec_format["nfft"]
    )

    # Construct a dataframe for each variable and add to the list of dataframes.
    for data_type, freq_column_names in column_names.items():
        df = raw_data.loc[:, ["time"] + freq_column_names]
        df.loc[:, freq_column_names] = df[freq_column_names] / (
            1000000.0 * frequency_resolution
        )

        rename_mapping = {name: index for index, name in enumerate(freq_column_names)}
        df = df.rename(rename_mapping, axis=1)
        data_types.append(data_type)
        dataframes.append(df)

    data = concat(dataframes, keys=data_types, names=["kind", "source_index"])
    data.reset_index(inplace=True)
    data.drop(columns="source_index", inplace=True)
    return data



from roguewave.timeseries_analysis.filtering import (
    sos_filter,
    spike_filter,
    exponential_delta_filter,
    exponential_filter,
    cumulative_filter,
    cumsum,
)
from roguewave import FrequencySpectrum
from roguewave.timeseries_analysis.time_integration import (
    integrate,
    complex_response,
    DEFAULT_N,
    DEFAULT_ORDER,
    cumulative_distance,
)
from numpy.typing import NDArray
from numpy import issubdtype, datetime64, ones
from roguewave.tools.time import datetime64_to_timestamp
from scipy.signal import butter
from numba import types
from numba.typed import Dict
from numbers import Number
from .welch import estimate_frequency_spectrum


DEFAULT_VELOCITY_PIPELINE = [
    ("spike", {"rejection_threshold": 9.81}),
    (
        "exponential_delta",
        {"smoothing_factor": 0.004, "maximum_gap_size": 3, "sampling_frequency": 2.5},
    ),
    ("integrate", {"order": DEFAULT_ORDER, "n": DEFAULT_N, "start_value": 0.0}),
    (
        "sos",
        {
            "direction": "forward",
            "sos": butter(4, 0.033, btype="high", output="sos", fs=2.5),
        },
    ),
]

DEFAULT_DISPLACEMENT_PIPELINE = [
    (
        "exponential_delta",
        {"smoothing_factor": 0.004, "maximum_gap_size": 3, "sampling_frequency": 2.5},
    ),
    (
        "sos",
        {
            "direction": "forward",
            "sos": butter(4, 0.04, btype="high", output="sos", fs=2.5),
        },
    ),
]

DEFAULT_SPOTTER_PIPELINE = DEFAULT_VELOCITY_PIPELINE


def pipeline(time: NDArray, signal: NDArray, stages=None) -> NDArray:
    if stages is None:
        stages = DEFAULT_SPOTTER_PIPELINE

    for index in range(0, len(stages)):
        value = stages[index]
        if isinstance(value, str):
            stages[index] = (value, {})

        elif isinstance(value, tuple):
            if value[1] is None:
                stages[index] = (value[0], {})

    if issubdtype(time.dtype, datetime64):
        time = datetime64_to_timestamp(time)

    for index, stage in enumerate(stages):
        signal = apply_filter(stage[0], time, signal, **stage[1])

    return signal


def spectral_pipeline(
    epoch_time: NDArray,
    x_or_u: NDArray,
    y_or_v: NDArray,
    z_or_w: NDArray,
    x_or_u_pipeline=None,
    y_or_v_pipeline=None,
    z_or_w_pipeline=None,
    latlon_input=False,
    window=None,
    segment_length_seconds=1800,
    sampling_frequency=2.5,
    spectral_window=None,
    response_correction=False,
) -> FrequencySpectrum:

    """
    Calculate the 1d frequency wave-spectrum including directional moments a1,b1,a2,b2 based on
    the raw input signals. The signals are processed according to the given pipeline prior to
    calculation of the co-spectra. If no pipelines are provided we use the default pipelines on
    Spotter which take as input: horizontal displacements x,y and vertical velocity w to produce
    the spectrum.

    :param time:
    :param x_or_u:
    :param y_or_v:
    :param z_or_w:
    :param x_or_u_pipeline:
    :param y_or_v_pipeline:
    :param z_or_w_pipeline:
    :param latlon_input:
    :param window:
    :param segment_length_seconds:
    :param spectral_window:
    :param response_correction:
    :return:
    """

    if latlon_input:
        x_or_u, y_or_v = cumulative_distance(x_or_u, y_or_v)

    if issubdtype(epoch_time.dtype, datetime64):
        epoch_time = datetime64_to_timestamp(epoch_time)

    if x_or_u_pipeline is None:
        x_or_u_pipeline = DEFAULT_DISPLACEMENT_PIPELINE
    if y_or_v_pipeline is None:
        y_or_v_pipeline = DEFAULT_DISPLACEMENT_PIPELINE
    if z_or_w_pipeline is None:
        z_or_w_pipeline = DEFAULT_VELOCITY_PIPELINE

    x = pipeline(epoch_time, x_or_u, x_or_u_pipeline)
    y = pipeline(epoch_time, y_or_v, y_or_v_pipeline)
    z = pipeline(epoch_time, z_or_w, z_or_w_pipeline)

    # Integration correction response funcitons
    response_functions = (
        get_response_correction(x_or_u_pipeline, apply=response_correction),
        get_response_correction(y_or_v_pipeline, apply=response_correction),
        get_response_correction(z_or_w_pipeline, apply=response_correction),
    )

    return estimate_frequency_spectrum(
        epoch_time,
        x,
        y,
        z,
        window,
        segment_length_seconds,
        sampling_frequency=sampling_frequency,
        spectral_window=spectral_window,
        response_functions=response_functions,
    )


def apply_filter(
    name: str, time_seconds: NDArray, signal: NDArray, **kwargs
) -> NDArray:
    """
    Apply the named filter to the signal and return the result.

    :param name: name of the filter to apply
    :param time_seconds: NDArray of time in seconds
    :param signal: NDArray of the signal (same length as time signal)
    :param kwargs: keyword filter options.
    :return: filtered signal
    """

    if name == "sos":
        if "direction" not in kwargs:
            kwargs["direction"] = "forward"
        return sos_filter(signal, **kwargs)

    elif name == "integrate":
        return integrate(
            time_seconds,
            signal,
            kwargs.get("order", DEFAULT_ORDER),
            kwargs.get("n", DEFAULT_N),
            kwargs.get("start_value", 0.0),
        )

    elif name == "spike":
        options = to_numba_kwargs(kwargs)
        return spike_filter(time_seconds, signal, options=options)

    elif name == "exponential_delta":
        options = to_numba_kwargs(kwargs)
        return exponential_delta_filter(time_seconds, signal, options=options)

    elif name == "cumulative":
        options = to_numba_kwargs(kwargs)
        return cumulative_filter(signal, options)

    elif name == "cumsum":
        options = to_numba_kwargs(kwargs)
        return cumsum(signal, options)

    elif name == "exponential":
        options = to_numba_kwargs(kwargs)
        return exponential_filter(time_seconds, signal, options)

    elif name == "identity":
        return signal

    else:
        raise Exception(f"unknown filter: {name}")


def to_numba_kwargs(kwargs):
    options = Dict.empty(key_type=types.unicode_type, value_type=types.float64)
    for k, v in kwargs.items():
        if isinstance(v, Number):
            options[k] = v

    return options


def get_response_correction(pipeline, apply=True):
    if not apply:
        return lambda f: ones(len(f), dtype="complex_")

    for stage in pipeline:
        if stage[0] == "integrate":
            return lambda f: 1 / complex_response(
                f, stage[1].get("order", DEFAULT_ORDER), stage[1].get("n", DEFAULT_N)
            )
    else:
        return lambda f: ones(len(f), dtype="complex_")



from roguewave.wavephysics.balance.dissipation import Dissipation
from roguewave.wavephysics.balance.generation import TWindInputType, WindGeneration
from roguewave.wavephysics.balance.balance import SourceTermBalance

from roguewave.wavephysics.balance.factory import (
    create_breaking_dissipation,
    breaking_parametrization,
    wind_parametrizations,
    create_wind_source_term,
    create_balance,
)



from .filecache import (
    filepaths,
    delete_cache,
    delete_files,
    create_cache,
    exists,
    delete_default,
    set_directive_function,
    remove_directive_function,
    get_cache,
    set,
)



"""
Contents: Wind Estimator

Copyright (C) 2022
Sofar Ocean Technologies

Authors: Pieter Bart Smit
"""
import typing
import numpy
from typing import Literal
from roguewave import FrequencySpectrum, FrequencyDirectionSpectrum, WaveSpectrum
from roguewave.wavephysics.balance import SourceTermBalance
from xarray import Dataset
from roguewave.wavephysics.roughness import charnock_roughness_length


_methods = Literal["peak", "mean"]
_charnock_parametrization = Literal["constant", "voermans15", "voermans16"]
_direction_convention = Literal[
    "coming_from_clockwise_north", "going_to_counter_clockwise_east"
]


def friction_velocity(
    spectrum: FrequencySpectrum,
    method: _methods = "peak",
    fmax=0.5,
    power=4,
    directional_spreading_constant=2.5,
    beta=0.012,
    grav=9.81,
    number_of_bins=20,
) -> Dataset:
    """

    :param spectrum:
    :param method:
    :param fmax:
    :param power:
    :param directional_spreading_constant:
    :param beta:
    :param grav:
    :param number_of_bins:
    :return:
    """
    e, a1, b1 = equilibrium_range_values(
        spectrum,
        method=method,
        fmax=fmax,
        power=power,
        number_of_bins=number_of_bins,
    )

    emean = 8.0 * numpy.pi**3 * e

    # Get friction velocity from spectrum
    friction_velocity_estimate = (
        emean / grav / directional_spreading_constant / beta / 4
    )

    # Estimate direction from tail
    direction = (180.0 / numpy.pi * numpy.arctan2(b1, a1)) % 360
    coords = {x: spectrum.dataset[x].values for x in spectrum.dims_space_time}
    return Dataset(
        data_vars={
            "friction_velocity": (spectrum.dims_space_time, friction_velocity_estimate),
            "direction": (spectrum.dims_space_time, direction),
        },
        coords=coords,
    )


def estimate_u10_from_spectrum(
    spectrum: WaveSpectrum,
    method: _methods = "peak",
    fmax=0.5,
    power=4,
    directional_spreading_constant=2.5,
    phillips_constant_beta=0.012,
    vonkarman_constant=0.4,
    grav=9.81,
    number_of_bins=20,
    direction_convention: _direction_convention = "going_to_counter_clockwise_east",
    **kwargs,
) -> Dataset:
    #
    # =========================================================================
    # Required Input
    # =========================================================================
    #
    # f              :: frequencies (in Hz)
    # E              :: Variance densities (in m^2 / Hz )
    #
    # =========================================================================
    # Output
    # =========================================================================
    #
    # U10            :: in m/s
    # Direction      :: in degrees clockwise from North (where wind is *coming from)
    #
    # =========================================================================
    # Named Keywords (parameters to inversion algorithm)
    # =========================================================================
    # Npower = 4     :: exponent for the fitted f^-N spectral tail
    # I      = 2.5   :: Philips Directional Constant
    # beta   = 0.012 :: Equilibrium Constant
    # Kapppa = 0.4   :: Von Karman constant
    # Alpha  = 0.012 :: Constant in estimating z0 from u* in Charnock relation
    # grav   = 9.81  :: Gravitational acceleration
    #
    # =========================================================================
    # Algorithm
    # =========================================================================
    #
    # 1) Find the part of the spectrum that best fits a f^-4 shape
    # 2) Estimate the Phillips equilibrium level "Emean" over that range
    # 3) Use Emean to estimate Wind speed (using Charnock and LogLaw)
    # 4) Calculate mean direction over equilibrium range

    if isinstance(spectrum, FrequencyDirectionSpectrum):
        spectrum_1d = spectrum.as_frequency_spectrum()
    else:
        spectrum_1d = spectrum

    # Get friction velocity from spectrum
    dataset = friction_velocity(
        spectrum_1d,
        method,
        fmax,
        power,
        directional_spreading_constant,
        phillips_constant_beta,
        grav,
        number_of_bins,
    )

    # Find z0 from Charnock Relation
    z0 = charnock_roughness_length(dataset.friction_velocity, **kwargs)

    if direction_convention == "coming_from_clockwise_north":
        dataset["direction"] = (270.0 - dataset["direction"]) % 360
    elif direction_convention == "going_to_counter_clockwise_east":
        pass
    else:
        raise ValueError(f"Unknown direectional convention: {direction_convention}")

    # Get the wind speed at U10 from loglaw
    return dataset.assign(
        {"u10": dataset.friction_velocity / vonkarman_constant * numpy.log(10.0 / z0)}
    )


def equilibrium_range_values(
    spectrum: FrequencySpectrum,
    method: _methods,
    fmax=1.25,
    power=4,
    number_of_bins=20,
) -> typing.Tuple[numpy.ndarray, numpy.ndarray, numpy.ndarray]:
    """

    :param spectrum:
    :param method:
    :param fmax:
    :param power:
    :param number_of_bins:
    :return:
    """

    if method == "peak":
        scaled_spec = spectrum.variance_density * spectrum.frequency**power
        scaled_spec = scaled_spec.fillna(0)
        indices = scaled_spec.argmax(dim="frequency")
        a1 = spectrum.a1.isel({"frequency": indices}).values
        b1 = spectrum.b1.isel({"frequency": indices}).values
        e = scaled_spec.isel({"frequency": indices}).values
        return e, a1, b1

    elif method == "mean":
        scaled_spec = spectrum.variance_density * spectrum.frequency**power

        # Find fmin/fmax
        fmin = 0
        i_min = numpy.argmin(numpy.abs(spectrum.frequency.values - fmin), axis=-1)
        i_max = numpy.argmin(numpy.abs(spectrum.frequency.values - fmax), axis=-1)
        nf = spectrum.number_of_frequencies

        i_max = i_max + 1 - number_of_bins
        i_max = numpy.max((i_min + 1, i_max))
        i_max = numpy.min((i_max, nf - number_of_bins))

        i_counter = 0
        shape = list(spectrum.shape())
        shape[-1] = i_max - i_min
        variance = numpy.zeros(shape) + numpy.inf

        #
        # Calculate the variance with respect to a running average mean of numberOfBins
        #
        for iFreq in range(i_min, i_max):
            #
            # Ensure we do not go out of bounds
            iiu = numpy.min((iFreq + number_of_bins, nf))

            # Ensure there are no 0 contributions (essentially no data)
            m = scaled_spec[..., iFreq:iiu].mean(dim="frequency")
            variance[..., i_counter] = ((scaled_spec[..., iFreq:iiu] - m) ** 2).mean(
                dim="frequency"
            ) / (m**2)
            i_counter = i_counter + 1
            #
        #
        i_min_variance = numpy.argmin(variance, axis=-1) + i_min

        e = numpy.zeros(i_min_variance.shape)
        a1 = numpy.zeros(i_min_variance.shape)
        b1 = numpy.zeros(i_min_variance.shape)

        index = numpy.array(list(range(0, spectrum.number_of_spectra)))
        index = numpy.unravel_index(index, i_min_variance.shape)
        i_min_variance = i_min_variance.flatten()

        for ii in range(0, number_of_bins):
            jj = numpy.clip(i_min_variance + ii, a_min=0, a_max=nf - 1 - number_of_bins)

            indexer = tuple([ind for ind in index] + [jj])

            e[index] += scaled_spec.values[indexer]
            a1[index] += spectrum.a1.values[indexer]
            b1[index] += spectrum.b1.values[indexer]
        fac = 1 / number_of_bins
        e *= fac
        a1 *= fac
        b1 *= fac

        return e, a1, b1


def estimate_u10_from_source_terms(
    spectrum: FrequencyDirectionSpectrum,
    balance: SourceTermBalance,
    time_derivative_spectrum=None,
    **kwargs,
) -> Dataset:
    # wind_direction = balance.dissipation.mean_direction_degrees(
    #     spectrum
    # )
    # dissipation_bulk = balance.dissipation.bulk_rate(spectrum)

    # Our first guess is the wind estimate based on the peak equilibrium range approximation
    guess = estimate_u10_from_spectrum(
        spectrum, "peak", direction_convention="going_to_counter_clockwise_east"
    )["u10"]

    u10_estimate = balance.windspeed_and_direction_from_spectra(
        guess, spectrum, time_derivative_spectrum=time_derivative_spectrum
    )
    #
    # u10_estimate = balance.generation.u10_from_bulk_rate(
    #         -dissipation_bulk, guess, wind_direction, spectrum)

    return u10_estimate



from .wavefields import bulk_parameters_partitions, DEFAULT_CONFIG_PARTITION_SPECTRA
from .wavefields import partition_spectra
from typing import List, Dict, Union
from datetime import timedelta
from pandas import DataFrame
from roguewave.wavespectra import FrequencySpectrum, FrequencyDirectionSpectrum
from typing import overload
from roguewave import logger

default_config = {
    'smoothInTime': False,
    'estimator': {
        'method': 'mem2',
        'numberOfDirections': 36,
        'frequencySmoothing': False,
        'smoothingLengthscale': 1
    },
    'partition_spectra':DEFAULT_CONFIG_PARTITION_SPECTRA
}


# -----------------------------------------------------------------------------
#                       Boilerplate Interfaces
# -----------------------------------------------------------------------------
@overload
def get_spectral_partitions_from_observations(
        spectra: Dict[str, FrequencySpectrum],
        minimum_duration: timedelta,
        config=None) -> Dict[str, List[FrequencyDirectionSpectrum]]: ...


@overload
def get_spectral_partitions_from_observations(
        spectra: FrequencySpectrum,
        minimum_duration: timedelta,
        config=None) -> List[FrequencyDirectionSpectrum]: ...


@overload
def get_bulk_partitions_from_observations(
        spectra: Dict[str, FrequencySpectrum],
        minimum_duration: timedelta,
        config=None,
        verbose=False) -> Dict[str, List[DataFrame]]: ...


@overload
def get_bulk_partitions_from_observations(
        spectra: FrequencySpectrum,
        minimum_duration: timedelta,
        config=None,
        verbose=False) -> List[DataFrame]: ...


# -----------------------------------------------------------------------------
#                       Implementation
# -----------------------------------------------------------------------------


def get_bulk_partitions_from_observations(
        spectra: Union[Dict[str, FrequencySpectrum], FrequencySpectrum],
        minimum_duration: timedelta,
        config=None,
        verbose=False) -> Union[Dict[str, List[DataFrame]], List[DataFrame]]:
    #
    spectral_partitions = get_spectral_partitions_from_observations(
        spectra, minimum_duration, config, verbose)

    if isinstance(spectra, dict):
        output = {}
        for key in spectral_partitions:
            output[key] = bulk_parameters_partitions(spectral_partitions[key])

    elif isinstance(spectra, list):
        output = bulk_parameters_partitions(spectral_partitions)
    else:
        raise Exception('Cannot process input')

    return output


def get_spectral_partitions_from_observations(
        spectra: Union[Dict[str, FrequencySpectrum], FrequencySpectrum],
        minimum_duration: timedelta,
        config=None) -> Union[
    Dict[str, List[FrequencyDirectionSpectrum]],
        List[FrequencyDirectionSpectrum]]:
    #

    if isinstance(spectra, dict):
        output = {}
        number = len(spectra)
        index = 0
        for key, item in spectra.items():
            index += 1
            if item is None:
                continue
            logger.info( f'{index:05d} out of {number:05d} spotter: {key}' )
            output[key] = partition_observations_spectra(item,
                                                         minimum_duration,
                                                         config,indent='    ')
    elif isinstance(spectra, list):
        output = partition_observations_spectra(spectra,
                                                minimum_duration,
                                                config)
    else:
        raise Exception('Cannot process input')

    return output


def partition_observations_spectra(spectra: FrequencySpectrum,
                                   minimum_duration: timedelta,
                                   config=None, indent='') -> List[
    FrequencyDirectionSpectrum]:
    if config:
        for key in config:
            assert key in default_config, f"{key} is not a valid conficuration entry"

        config = default_config | config
    else:
        config = default_config

    # Step 1: Pre-Processing
    # :
    # Prior to constructing spectra - smoothing the wave field can help create
    # more stable results.

    if config['smoothInTime']:
        pass

    # Step 2: Create 2D wavefields from 1D spectra using a spectral estimator

    logger.info( indent + ' - Create 2D Spectra')
    spectra2D = spectra.as_frequency_direction_spectrum(
        number_of_directions=36)
    return partition_spectra(
        spectra2D, minimum_duration, config=config['partition_spectra'],indent=indent)





from numpy.typing import ArrayLike
from xarray import DataArray
from numpy import inf, isfinite, abs, nansum, nanmax
import xarray
import numpy
from typing import TypeVar, Callable, List
from roguewave.log import logger
from logging import DEBUG, INFO
from dataclasses import dataclass

_T = TypeVar("_T", ArrayLike, DataArray)

_iteration_depth = 0


@dataclass()
class Configuration:
    atol: float = 1e-4
    rtol: float = 1e-4
    max_iter: int = 100
    aitken_acceleration: bool = True
    fraction_of_points: float = 1
    error_if_not_converged: bool = False
    numerical_derivative_stepsize: float = 1e-4
    use_numba: bool = True


def _log(msg, level):
    if _iteration_depth > 1:
        # Sub iterations are set to level debug
        level = DEBUG
    logger.log(level, msg)


def fixed_point_iteration(
    function: Callable[[List[_T]], _T],
    guess: _T,
    bounds=(-inf, inf),
    configuration: Configuration = None,
    caller: str = None,
) -> _T:
    """
    Fixed point iteration on a vector function. We want to solve the parallal problem x=F(x) where x is a vector. Instead
    of looping over each problem and solving them individualy using e.g. scipy solvers, we gain some efficiency by
    evaluating F in parallel, and doing the iteration ourselves. Only worthwhile if F is the expensive part and/or x
    is large.

    :param function:
    :param guess:
    :param max_iter:
    :param atol:
    :param rtol:
    :param caller:
    :return:
    """

    # Get the current identation level. This is only used for displaying messages in logs.
    global _iteration_depth
    whitespace = _iteration_depth * "\t"
    # Increase identationm level in case of any recursive calling of solvers.
    _iteration_depth += 1

    iterates = [guess, guess, guess]

    if configuration is None:
        configuration = Configuration()

    guess = iterates[2]
    mask_finite_guess_points = numpy.isfinite(guess)
    number_of_active_points = nansum(mask_finite_guess_points)

    is_dataarray = isinstance(guess, DataArray)
    if is_dataarray:
        where = xarray.where
    else:
        where = numpy.where

    if caller is None:
        caller = "unknown"

    msg = f"{whitespace}Starting solver"
    _log(msg, INFO)

    converged = numpy.zeros(guess.shape, dtype="bool")
    for current_iteration in range(1, configuration.max_iter + 1):
        # Update iterate
        aitken_step = configuration.aitken_acceleration and current_iteration % 3 == 0
        if aitken_step:
            # Every third step do an aitken acceleration step- if requested
            ratio = (iterates[2] - iterates[1]) / (iterates[1] - iterates[0])
            ratio = where(numpy.isfinite(ratio) & (ratio != 1.0), ratio, 0.0)

            next_iterate = iterates[2] + ratio / (1.0 - ratio) * (
                iterates[2] - iterates[1]
            )
        else:
            next_iterate = function(iterates[2])

        # Bounds check
        if isfinite(bounds[0]):
            next_iterate = where(
                (next_iterate <= bounds[0]),
                (bounds[0] - iterates[2]) * 0.5 + iterates[2],
                next_iterate,
            )

        if isfinite(bounds[1]):
            next_iterate = where(
                (next_iterate > bounds[1]),
                (bounds[1] - iterates[2]) * 0.5 + iterates[2],
                next_iterate,
            )

        # Roll the iterates, make the last entry the latest estimate
        iterates.append(iterates.pop(0))
        iterates[2] = next_iterate

        # Convergence check
        absolute_difference = abs(iterates[2] - iterates[1])

        scale = numpy.maximum(abs(iterates[1]), configuration.atol)
        relative_difference = absolute_difference / scale
        converged[:] = (absolute_difference < configuration.atol) & (
            relative_difference < configuration.rtol
        )

        max_abs_error = nanmax(absolute_difference)
        max_rel_error = nanmax(relative_difference)
        percent_converged = nansum(converged) / number_of_active_points * 100

        number_of_points_converged = nansum(converged)

        msg = f"{whitespace} - Iteration {current_iteration} (max {configuration.max_iter}), convergence in {percent_converged:.2f} % points"
        msg = (
            msg
            + f"max abs. errors: {max_abs_error:.2e} (atol: {configuration.atol}), max rel. error: {max_rel_error:.2e} (rtol: {configuration.rtol})"
        )
        if number_of_points_converged == number_of_active_points and not aitken_step:
            _log(msg, INFO)
            break

        elif (
            number_of_points_converged
            >= number_of_active_points * configuration.fraction_of_points
        ) and not aitken_step:
            _log(msg, INFO)
            break

        else:
            _log(msg, INFO)

    else:
        msg = f"{whitespace}No convergence after {configuration.max_iter}"
        if configuration.error_if_not_converged:
            raise ValueError(msg)

        else:
            iterates[2][~converged] = numpy.nan
            _log(msg, INFO)

    # Reduce identation level in logging
    _iteration_depth -= 1
    return iterates[2]



from roguewave import FrequencyDirectionSpectrum
from abc import ABC
from numba import njit, types
from numba.typed import Dict as NumbaDict
from typing import MutableMapping


class SourceTerm(ABC):
    def __init__(self, parameters):
        self._parameters = (
            parameters if parameters is not None else self.default_parameters()
        )

    @property
    def parameters(self) -> NumbaDict:
        return _numba_parameters(**self._parameters)

    def spectral_grid(self, spectrum: FrequencyDirectionSpectrum):
        return _spectral_grid(
            spectrum.radian_frequency.values,
            spectrum.radian_direction.values,
            spectrum.frequency_step.values,
            spectrum.direction_step.values,
        )

    @staticmethod
    def default_parameters() -> MutableMapping:
        return {}

    @parameters.setter
    def parameters(self, parameters):
        self._parameters = parameters


@njit()
def _spectral_grid(radian_frequency, radian_direction, frequency_step, direction_step):
    return {
        "radian_frequency": radian_frequency,
        "radian_direction": radian_direction,
        "frequency_step": frequency_step,
        "direction_step": direction_step,
    }


def _numba_parameters(**kwargs):
    _dict = NumbaDict.empty(key_type=types.unicode_type, value_type=types.float64)
    for key in kwargs:
        _dict[key] = kwargs[key]

    return _dict



"""
Contents: Spectral estimators

Copyright (C) 2022
Sofar Ocean Technologies

Authors: Pieter Bart Smit
======================

Spectral estimators that can be used to create a 2D spectrum from buoy
observations

Classes:

- `None

Functions:

- `mem`, maximum entrophy method
- `mem2`, ...

How To Use This Module
======================
(See the individual functions for details.)
"""
from roguewave.wavespectra.estimators.mem2 import mem2
from roguewave.wavespectra.estimators.mem import mem

# from roguewave.wavespectra.estimators.loglikelyhood import log_likelyhood
from roguewave.wavespectra.estimators.estimate import (
    estimate_directional_distribution,
    estimate_directional_spectrum_from_moments,
    Estimators,
)



from roguewave.wavephysics.fluidproperties import (
    AIR,
    WATER,
    FluidProperties,
    GRAVITATIONAL_ACCELERATION,
)
from roguewave import FrequencyDirectionSpectrum, SPECTRAL_DIMS
from roguewave.wavephysics.generation import WindGeneration, TWindInputType
from numpy import cos, pi, log, tanh, sqrt
from xarray import DataArray, where
from roguewave import integrate_spectral_data
from scipy.special import kelvin


class SwellDissipation(WindGeneration):
    def __init__(
        self,
        gravitational_acceleration=GRAVITATIONAL_ACCELERATION,
        swell_dissipation_coefficients=None,
        **kwargs,
    ):
        super(WindGeneration, self).__init__(**kwargs)
        self.gravitational_acceleration = gravitational_acceleration
        if swell_dissipation_coefficients is None:
            swell_dissipation_coefficients = {
                "s0": 3,
                "s1": 0.8,
                "s2": -0.018,
                "s3": 0.015,
                "rz0": 0.04,
                "laminar_coeficient_cds": 1.2,
                "dimensional_critical_reynolds_number": 2e5,
            }
        self.swell_dissipation_coefficients = swell_dissipation_coefficients

    def rate(
        self,
        speed: DataArray,
        direction: DataArray,
        spectrum: FrequencyDirectionSpectrum,
        roughness_length: DataArray = None,
        wind_speed_input_type: TWindInputType = "u10",
        air: FluidProperties = AIR,
        water: FluidProperties = WATER,
        memoized=None,
    ) -> DataArray:

        if wind_speed_input_type == "u10":
            return self.rate_U10(
                speed, direction, spectrum, roughness_length, air, water, memoized
            )

        elif wind_speed_input_type in ["ustar", "friction_velocity"]:
            return self.rate_friction_velocity(
                speed, direction, spectrum, roughness_length, air, water, memoized
            )

        else:
            raise ValueError(
                f"Unknown input type {wind_speed_input_type}, "
                f"has to be one of: 'u10','friction_velocity','ustar'"
            )

    def rate_U10(
        self,
        speed: DataArray,
        direction: DataArray,
        spectrum: FrequencyDirectionSpectrum,
        roughness_length: DataArray = None,
        air: FluidProperties = AIR,
        water: FluidProperties = WATER,
        memoized=None,
    ) -> DataArray:

        friction_velocity = air.vonkarman_constant * speed / log(10 / roughness_length)
        return self.rate_friction_velocity(
            friction_velocity,
            direction,
            spectrum,
            roughness_length,
            air,
            water,
            memoized,
        )

    def rate_friction_velocity(
        self,
        speed: DataArray,
        direction: DataArray,
        spectrum: FrequencyDirectionSpectrum,
        roughness_length: DataArray = None,
        air: FluidProperties = AIR,
        water: FluidProperties = WATER,
        memoized=None,
    ) -> DataArray:

        memoized = memoized if memoized is not None else {}
        if "wave_speed" not in memoized:
            memoized["wave_speed"] = spectrum.wave_speed()

        if "wavenumber" not in memoized:
            memoized["wavenumber"] = spectrum.wavenumber

        if "significant_wave_height" not in memoized:
            memoized["significant_wave_height"] = spectrum.significant_waveheight

        if "significant_orbital_velocity" not in memoized:
            memoized["significant_orbital_velocity"] = st4_significant_orbital_velocity(
                spectrum.variance_density,
                spectrum.radian_frequency,
                memoized["wavenumber"],
                spectrum.depth,
            )

        if "wave_reynolds_number" not in memoized:
            memoized["wave_reynolds_number"] = st4_wave_reynolds_number(
                memoized["significant_orbital_velocity"],
                memoized["significant_wave_height"] / 2,
                air,
            )

        if "mutual_angle" not in memoized:
            memoized["mutual_angle"] = (
                ((spectrum.direction - direction + 180.0) % 360.0 - 180.0) * pi / 180
            )

        critical_wave_reynolds_number = st4_crictical_reynolds_number(
            self.swell_dissipation_coefficients, memoized["significant_wave_height"]
        )

        swell_dissipation = st4_swell_dissipation(
            speed,
            memoized["mutual_angle"],
            spectrum.variance_density,
            roughness_length,
            memoized["significant_wave_height"] / 2,
            memoized["wave_reynolds_number"],
            critical_wave_reynolds_number,
            memoized["wavenumber"],
            spectrum.radian_frequency,
            memoized["significant_orbital_velocity"],
            self.swell_dissipation_coefficients,
            self.gravitational_acceleration,
            air,
            water,
        )
        return swell_dissipation


def st4_wave_reynolds_number(
    significant_orbital_velocity, significant_amplitude, air: FluidProperties = AIR
):
    return (
        4
        * significant_orbital_velocity
        * significant_amplitude
        / air.kinematic_viscosity
    )


def st4_significant_orbital_velocity(
    variance_density, radian_frequency, wavenumber, depth
):
    return 2 * sqrt(
        integrate_spectral_data(
            variance_density * radian_frequency**2 * tanh(wavenumber * depth),
            dims=SPECTRAL_DIMS,
        )
    )


def st4_crictical_reynolds_number(
    swell_dissipation_coefficients, significant_wave_height
):
    return (
        swell_dissipation_coefficients["dimensional_critical_reynolds_number"]
        / significant_wave_height
    )


def st4_swell_dissipation(
    speed,
    mutual_angle,
    variance_density,
    roughness,
    significant_amplitude,
    wave_reynolds_number,
    critical_reynolds_number,
    wavenumber,
    angular_frequency,
    significant_orbital_velocity,
    swell_dissipation_coefficients,
    gravitational_acceleration,
    air: FluidProperties = AIR,
    water: FluidProperties = WATER,
):
    ratio = air.density / water.density

    criterion = wave_reynolds_number <= critical_reynolds_number
    laminar_coeficient = swell_dissipation_coefficients["laminar_coeficient_cds"]
    laminar = -(
        laminar_coeficient
        * ratio
        * 2
        * wavenumber
        * sqrt(2 * angular_frequency * air.kinematic_viscosity)
    )

    swell_dissipation_factor = st4_swell_dissipation_factor(
        speed,
        significant_orbital_velocity,
        roughness,
        significant_amplitude,
        mutual_angle,
        swell_dissipation_coefficients,
        water,
    )

    turbulent = -(
        ratio
        * 16
        * swell_dissipation_factor
        * angular_frequency**2
        * significant_orbital_velocity
        / gravitational_acceleration
    )

    return where(criterion, laminar, turbulent) * variance_density


def st4_swell_dissipation_factor(
    speed,
    significant_orbital_velocity,
    roughness,
    significant_amplitude,
    mutual_angle,
    swell_dissipation_coefficients,
    water: FluidProperties,
):
    dissipation_factor_grant_maddsen = st4_dissipation_factor_grant_maddsen(
        roughness, significant_amplitude, swell_dissipation_coefficients, water
    )

    return swell_dissipation_coefficients["s1"] * (
        dissipation_factor_grant_maddsen
        + (
            swell_dissipation_coefficients["s3"]
            + swell_dissipation_coefficients["s2"] * cos(mutual_angle)
        )
        * speed
        / significant_orbital_velocity
    )


def st4_dissipation_factor_grant_maddsen(
    roughness,
    significant_amplitude,
    swell_dissipation_coefficients,
    water: FluidProperties = WATER,
):
    dimensionless_roughness = (
        swell_dissipation_coefficients["rz0"] * roughness / significant_amplitude
    )

    kel = kelvin(2 * sqrt(dimensionless_roughness))[1]
    return water.vonkarman_constant**2 / (abs(kel) ** 2 * 2)



from json import dumps

_formats = {
    "GPS": {
        "sampling_interval_seconds": 0.4,
        "time_column": "GPS_Epoch_Time(s)",
        "pattern": "????_GPS.csv",
        "ragged": False,
        "dropna": True,
        "drop": [],
        "columns": [
            {
                "name": "millis",
                "dtype": "float64",
            },
            {
                "name": "GPS_Epoch_Time(s)",
                "dtype": "float64",
            },
            {
                "name": "lat(deg)",
                "dtype": "float64",
            },
            {
                "name": "lat(min*1e5)",
                "dtype": "float64",
            },
            {
                "name": "long(deg)",
                "dtype": "float64",
            },
            {
                "name": "long(min*1e5)",
                "dtype": "float64",
            },
            {
                "name": "el(m)",
                "dtype": "float64",
            },
            {
                "name": "SOG(mm_s)",
                "dtype": "float64",
            },
            {
                "name": "COG(deg*1000)",
                "dtype": "float64",
            },
            {
                "name": "vert_vel(mm_s)",
                "dtype": "float64",
            },
        ],
    },
    "FLT": {
        "sampling_interval_seconds": 0.4,
        "time_column": "GPS_Epoch_Time(s)",
        "pattern": "????_FLT.csv",
        "ragged": False,
        "dropna": True,
        "drop": ["flag"],
        "columns": [
            {
                "name": "millis",
                "dtype": "float64",
            },
            {
                "name": "GPS_Epoch_Time(s)",
                "dtype": "float64",
            },
            {
                "name": "outx(mm)",
                "dtype": "float64",
            },
            {
                "name": "outy(mm)",
                "dtype": "float64",
            },
            {
                "name": "outz(mm)",
                "dtype": "float64",
            },
            {
                "name": "flag",
                "dtype": "str",
            },
        ],
    },
    "LOC": {
        "sampling_interval_seconds": 60,
        "time_column": "GPS_Epoch_Time(s)",
        "pattern": "????_LOC.csv",
        "ragged": False,
        "dropna": True,
        "drop": [],
        "columns": [
            {
                "name": "GPS_Epoch_Time(s)",
                "dtype": "float64",
            },
            {
                "name": "lat(deg)",
                "dtype": "float64",
            },
            {
                "name": "lat(min*1e5)",
                "dtype": "float64",
            },
            {
                "name": "long(deg)",
                "dtype": "float64",
            },
            {
                "name": "long(min*1e5)",
                "dtype": "float64",
            },
        ],
    },
    "TIME": {
        "sampling_interval_seconds": 60,
        "time_column": "GPS_Epoch_Time(s)",
        "pattern": "????_LOC.csv",
        "ragged": False,
        "dropna": True,
        "drop": [],
        "columns": [
            {
                "name": "GPS_Epoch_Time(s)",
                "dtype": "float64",
            },
            {
                "name": "lat(deg)",
                "dtype": "float64",
            },
            {
                "name": "lat(min*1e5)",
                "dtype": "float64",
            },
            {
                "name": "long(deg)",
                "dtype": "float64",
            },
            {
                "name": "long(min*1e5)",
                "dtype": "float64",
            },
        ],
    },
    "GMN": {
        "sampling_interval_seconds": 10,
        "time_column": "millis",
        "pattern": "????_GMN.csv",
        "ragged": True,
        "dropna": False,
        "drop": [],
        "columns": [
            {
                "name": "millis",
                "dtype": "int64",
            },
            {
                "name": "noisePerMs",
                "dtype": "float64",
            },
            {
                "name": "agcCnt",
                "dtype": "float64",
            },
            {
                "name": "aStatus",
                "dtype": "float64",
            },
            {
                "name": "aPower",
                "dtype": "float64",
            },
            {
                "name": "jamInd",
                "dtype": "float64",
            },
            {
                "name": "jamStat",
                "dtype": "float64",
            },
            {
                "name": "ofsI",
                "dtype": "float64",
            },
            {
                "name": "magI",
                "dtype": "float64",
            },
            {
                "name": "ofsQ",
                "dtype": "float64",
            },
            {
                "name": "magQ",
                "dtype": "float64",
            },
            {
                "name": "VN(mm_s)",
                "dtype": "float64",
            },
            {
                "name": "VE(mm_s)",
                "dtype": "float64",
            },
            {
                "name": "VD(mm_s)",
                "dtype": "float64",
            },
            {
                "name": "avgSignalStrength",
                "dtype": "float64",
            },
            {
                "name": "SVs_used",
                "dtype": "float64",
            },
            {
                "name": "SVs_tracked",
                "dtype": "float64",
            },
            {
                "name": "signal_strengths_0",
                "dtype": "float64",
            },
            {
                "name": "signal_strengths_1",
                "dtype": "float64",
            },
            {
                "name": "signal_strengths_2",
                "dtype": "float64",
            },
            {
                "name": "signal_strengths_3",
                "dtype": "float64",
            },
            {
                "name": "signal_strengths_4",
                "dtype": "float64",
            },
            {
                "name": "signal_strengths_5",
                "dtype": "float64",
            },
            {
                "name": "signal_strengths_6",
                "dtype": "float64",
            },
            {
                "name": "signal_strengths_7",
                "dtype": "float64",
            },
            {
                "name": "signal_strengths_8",
                "dtype": "float64",
            },
            {
                "name": "signal_strengths_9",
                "dtype": "float64",
            },
            {
                "name": "signal_strengths_10",
                "dtype": "float64",
            },
            {
                "name": "signal_strengths_11",
                "dtype": "float64",
            },
            {
                "name": "signal_strengths_12",
                "dtype": "float64",
            },
            {
                "name": "signal_strengths_13",
                "dtype": "float64",
            },
            {
                "name": "signal_strengths_14",
                "dtype": "float64",
            },
            {
                "name": "signal_strengths_15",
                "dtype": "float64",
            },
            {
                "name": "signal_strengths_16",
                "dtype": "float64",
            },
            {
                "name": "signal_strengths_17",
                "dtype": "float64",
            },
            {
                "name": "signal_strengths_18",
                "dtype": "float64",
            },
            {
                "name": "signal_strengths_19",
                "dtype": "float64",
            },
        ],
    },
    "BARO_RAW": {
        "sampling_interval_seconds": 0.2,
        "time_column": "timestamp (ticks/UTC)",
        "pattern": "????_BARO_RAW.csv",
        "ragged": False,
        "dropna": True,
        "drop": [],
        "columns": [
            {
                "name": "timestamp (ticks/UTC)",
                "dtype": "float64",
            },
            {
                "name": "temperature (C)",
                "dtype": "float64",
            },
            {
                "name": "pressure (mbar)",
                "dtype": "float64",
            },
        ],
    },
    "BARO": {
        "sampling_interval_seconds": 60,
        "time_column": "timestamp (ticks/UTC)",
        "pattern": "????_BARO.csv",
        "ragged": False,
        "dropna": True,
        "drop": [],
        "columns": [
            {
                "name": "timestamp (ticks/UTC)",
                "dtype": "float64",
            },
            {
                "name": "pressure (mbar)",
                "dtype": "float64",
            },
        ],
    },
    "SST": {
        "sampling_interval_seconds": 60,
        "time_column": "timestamp (ticks/UTC)",
        "pattern": "????_SST.csv",
        "ragged": False,
        "dropna": True,
        "drop": [],
        "columns": [
            {
                "name": "timestamp (ticks/UTC)",
                "dtype": "float64",
            },
            {
                "name": "temperature (C)",
                "dtype": "float64",
            },
        ],
    },
    "RAINDB": {
        "sampling_interval_seconds": 300,
        "time_column": "timestamp (ticks/UTC)",
        "pattern": "????_SST.csv",
        "ragged": False,
        "dropna": True,
        "drop": [],
        "columns": [
            {
                "name": "timestamp (ticks/UTC)",
                "dtype": "float64",
            },
            {
                "name": "level (dB)",
                "dtype": "float64",
            },
        ],
    },
}

if __name__ == "__main__":
    for k, v in _formats.items():
        with open(f"{k}.json", "wt") as file:
            str = dumps(v, indent=4)
            file.write(str)



from roguewave.wavewatch3.restart_file import RestartFile
from roguewave.wavewatch3.restart_file_time_stack import RestartFileTimeStack
from roguewave.wavewatch3.io import (
    clone_restart_file,
    open_restart_file,
    write_partial_restart_file,
    write_restart_file,
    reassemble_restart_file_from_parts,
    open_restart_file_stack,
)

from roguewave.wavewatch3.resources import create_resource
from roguewave.wavewatch3.restart_file_metadata import read_header, MetaData
from roguewave.wavewatch3.model_definition import read_model_definition
from roguewave.wavewatch3.grid_tools import Grid, unpack_ww3_data



import numpy
from datetime import timedelta
from typing import List, Union, Dict, overload
from pandas import DataFrame
from roguewave import logger

from roguewave.wavespectra import (
    FrequencyDirectionSpectrum, WaveSpectrum)
from .partitioning import is_neighbour, default_partition_config, \
    partition_spectrum
from multiprocessing import cpu_count, get_context

DEFAULT_CONFIG_PARTITION_SPECTRA = {
    'partitionConfig': default_partition_config,
    'parallel': False,  # DOES NOT WORK
    'fieldFiltersettings': {
        'filter': True,
        'maxDeltaPeriod': 2,
        'maxDeltaDirection': 40,
        'maxDeltaWaveHeight': 1.5,
        'minDeltaPeriod': 1,
        'minDeltaDirection': 15,
        'minDeltaWaveHeight': 0.5
    }
}


class Node():
    def __init__(self,
                 spectrum_index: int,
                 correlation_with_parent: float,
                 spectrum: Union[FrequencyDirectionSpectrum, None]):

        self.indices = [spectrum_index]
        self.spectra = [spectrum]
        self.children = []  # type: List["Node"]
        self.correlation_with_previous = [correlation_with_parent]
        self.favored_child = None  # type: Union["Node",None]
        self.parent = None  # type: Union["Node",None]
        self._staged = []

    def stage(self, index, spectrum, correlation):
        self._staged.append((index, spectrum, correlation))

    def commit(self) -> List["Node"]:

        # if no partitions are stages- nothing to be done, return empty list
        # (signifying no nodes were added)
        if not self._staged:
            return []

        # Partitions are staged...
        if len(self._staged) > 1:
            # if multiple partitions are staged- this is a branch point, add
            # children
            for index, spectrum, correlation in self._staged:
                child = Node(index,
                             correlation,
                             spectrum)
                self.add_child(child)

            self._staged = []

            # we return the children- these are now active to continue the path
            return self.children
        else:
            #
            # Single partition, just add to the current spectra
            self.add_spectrum(*self._staged[0])
            self._staged = []

            # return self- is still active.
            return [self]

    def add_spectrum(self, index, spectrum, correlation):
        self.spectra.append(spectrum)
        self.correlation_with_previous.append(correlation)
        self.indices.append(index)

    def add_child(self, child: "Node"):
        #
        child.parent = self
        self.children.append(child)
        self.update_favored_child()

    @property
    def duration(self):
        delta = self.spectra[-1].time - self.spectra[0].time
        if self.parent is None:
            return delta

        if self.parent.favored_child is self:
            delta = delta + self.parent.duration
        return delta

    def prune(self, min_duration: timedelta):
        if self.duration > min_duration:
            return False
        else:
            self.parent.remove_child(self)
            return True

    def update_favored_child(self):
        self.favored_child = None
        corr = 0
        for candidate_child in self.children:
            if candidate_child.correlation_with_previous[0] >= corr:
                corr = candidate_child.correlation_with_previous[0]
                self.favored_child = candidate_child

    def remove_child(self, child: "Node"):
        assert not child.children
        for index, sibling in enumerate(self.children):
            if sibling is child:
                child_index = index
                break
        else:
            raise Exception('not a child')

        self.children.pop(child_index)
        self.merge(child.indices, child.spectra)
        child.parent = None

    def merge(self, indices: List[int],
              spectra: List[FrequencyDirectionSpectrum]):
        if not len(indices):
            return

        if not self.children:
            self.merge_spectra(indices, spectra)
        else:
            if indices[0] <= self.indices[-1]:
                istart = self.indices.index(indices[0])
                length = min(len(self.indices) - istart, len(indices))

            else:
                length = 0
                assert indices[0] == self.indices[-1] + 1

            self.merge_spectra(indices[0:length], spectra[0:length])
            if length == len(spectra):
                return

            number_of_neighbourpoints = numpy.zeros(len(self.children),
                                                    dtype='float64')
            for ii, child in enumerate(self.children):
                number_of_neighbourpoints[ii] = is_neighbour(spectra[length],
                                                             child.spectra[0])

            index_of_child_to_merge_into = numpy.argmax(
                number_of_neighbourpoints)

            self.children[index_of_child_to_merge_into].merge(indices[length:],
                                                              spectra[length:])

        # Since correlation between children and parents has changed, update
        # the correlations.
        self.update_favored_child()

    def merge_spectra(self, indices: List, spectra):
        if not indices:
            return

        if indices[0] > self.indices[-1]:
            assert indices[0] == self.indices[-1] + 1
            istart = self.indices[-1] + 1
        else:
            istart = self.indices.index(indices[0])

        jj = istart
        for index, spectrum in zip(indices, spectra):
            #
            if index > self.indices[-1]:
                correlation = correlate(self.spectra[-1], spectrum)
                self.add_spectrum(index, spectrum, correlation)
            else:
                self.spectra[jj] = self.spectra[jj] + spectrum
                if jj == 0:
                    correlation = correlate(self.parent.spectra[-1],
                                            self.spectra[jj])
                else:
                    correlation = correlate(self.spectra[jj - 1],
                                            self.spectra[jj])
                self.correlation_with_previous[jj] = correlation
            jj = jj + 1


def correlate(source: FrequencyDirectionSpectrum,
              target: FrequencyDirectionSpectrum):
    source = source.spectral_values.values
    target = target.spectral_values.values
    return numpy.nansum(source * target)


def correlate_spectra_with_nodes(source_spectra:
                        List[FrequencyDirectionSpectrum],
                                 nodes: List[Node]) -> numpy.array:
    correlations = numpy.zeros((len(source_spectra), len(nodes)))

    for i_spec, source_spectrum in enumerate(source_spectra):
        for i_node, target_node in enumerate(nodes):
            correlations[i_spec, i_node] = correlate(source_spectrum,
                                                     target_node.spectra[-1])
    return correlations


def create_graph(
        partitions_list: List[Dict[int, FrequencyDirectionSpectrum]],
        min_duration):
    label = list(partitions_list[0].keys())[0]
    empty = partitions_list[0][label].copy(deep=True)
    empty['variance_density'] = 0 * empty['variance_density']
    root = Node(-1, 0, empty)
    for label, spectrum in partitions_list[0].items():
        root.add_child(Node(0, 1, spectrum))

    active_nodes = root.children
    for index in range(1, len(partitions_list)):

        new_active_set = []
        labels = list(partitions_list[index].keys())
        partitions = [partitions_list[index][label] for label in labels]

        # Correlate all active nodes with new candidate partitions
        correlations = correlate_spectra_with_nodes(partitions, active_nodes)

        # get the strongest correlations between new partitions and previous
        # nodes
        new_to_previous = numpy.argmax(correlations, axis=1)

        # connect new partition to previous nodes. We first stage the commits
        # so we now how many partitions are connected to a previous node
        for new_index, max_index in enumerate(new_to_previous):
            active_nodes[max_index].stage(
                index,
                partitions[new_index],
                correlations[new_index, max_index]
            )

        for node in active_nodes:
            if nodes := node.commit():
                new_active_set += nodes
            else:
                # No spectra have been added - this is the end for this partition
                pass
        active_nodes = new_active_set

    tree_prune(root, min_duration)
    return root


def tree_prune(root: Node, min_duration: timedelta):
    """
    Remove branches from the tree that contain only a few speectra.
    :param root:
    :param min_duration:
    :return:
    """

    iteration = 0
    logger.debug(f"Pruning the tree")
    while branch_prune(root, min_duration):
        iteration += 1
        logger.debug(f"\t Iteration {iteration:04d}")
    return


def branch_prune(root: Node, min_duration: timedelta) -> bool:
    pruned = False
    if root.children:
        for child in root.children:
            pruned = pruned or branch_prune(child, min_duration)
    else:
        if root.parent:
            pruned = root.prune(min_duration)
    return pruned


def wave_fields_from(root: Node, primary_field=None) -> List[
    List[FrequencyDirectionSpectrum]]:
    """
    Create different swell/sea fields based on a given connectivity tree.

    :param root: root of the tree structure
    :param primary_field: memory parameter in the recursive construction.
    :return:
    """

    # Initialize the output fields of this root to eempty
    fields = []

    # If we do not have a parent we are the root of three- each of the children
    # is considered a valid field>
    if root.parent is None:
        for child in root.children:
            # Recurisively call method on children. Out output field from children
            # to current field
            fields += wave_fields_from(child, [])
    else:
        # add spectra of the current node to the wave field
        for spectrum in root.spectra:
            primary_field.append(spectrum)

        # if we have children, add their spectra too
        if root.children:
            for child in root.children:
                if child is root.favored_child:
                    # if this is the "favored" child this continuous the primary partition
                    fields += wave_fields_from(child, primary_field)
                else:
                    # if this is not the favored child, this is a new field.
                    fields += wave_fields_from(child, [])
        else:
            # if no children, this is the end. Return the primary partition as
            # a field.
            fields += [primary_field]

    return fields


def are_different_fields( spec1:FrequencyDirectionSpectrum,
                          spec2:FrequencyDirectionSpectrum,config ):
    def ramp( value, min, max ):
        if value < min:
            return 0
        elif value > max:
            return 1
        else:
            return (value - min) / (max - min)

    delta_waveheight = numpy.abs( spec1.hm0()- spec2.hm0())
    delta_period = numpy.abs(spec1.tm01() - spec2.tm01())
    delta_direction = numpy.abs(
        (spec1.mean_direction() - spec2.mean_direction() + 180) % 360 - 180
    )

    return \
        ramp(delta_waveheight,config['minDeltaWaveHeight'],config['maxDeltaWaveHeight']) \
    +   ramp(delta_period,config['minDeltaPeriod'],config['maxDeltaPeriod']) \
    +   ramp(delta_direction,config['minDeltaDirection'],config['maxDeltaDirection']) \
    > 1


def filter_field(field: List[FrequencyDirectionSpectrum],
                 min_duration: timedelta,
                 config):
    new_fields = [[]]  # type: List[List[FrequencyDirectionSpectrum]]
    current_field = 0
    for index, spec in enumerate(field):
        if index == 0:
            new_fields[current_field].append(spec)
            continue

        previous_field = field[index - 1]
        # delta_period = numpy.abs(spec.tm01() - previous_field.tm01())
        # delta_direction = numpy.abs(
        #     (spec.mean_direction() - previous_field.mean_direction() + 180) % 360 - 180
        # )
        if are_different_fields(spec,previous_field,config): #delta_direction > config['maxDeltaDirection'] or delta_period > config['maxDeltaPeriod']:
            current_field += 1
            new_fields.append([])
        new_fields[current_field].append(spec)

    to_drop = []
    for index, new_field in enumerate(new_fields):
        if new_field[-1].time - new_field[0].time < min_duration:
            to_drop.append(index)

    for index in sorted(to_drop, reverse=True):
        new_fields.pop(index)
    return new_fields


def filter_fields(fields: List[List[FrequencyDirectionSpectrum]],
                  min_duration: timedelta,config):
    new_fields = []
    for field in fields:
        new_fields += filter_field(field, min_duration, config)
    return new_fields


def extract_bulk_parameter(parameter, spectra: WaveSpectrum):
    temp = getattr(spectra, parameter)
    if callable(temp):
        output = temp()
    else:
        output = temp
    return output


def bulk_parameters_partitions(
        partitions: List[FrequencyDirectionSpectrum]) -> List[DataFrame]:
    bulk = []
    for label, partition in enumerate(partitions):
        df = DataFrame()
        for variable in WaveSpectrum.bulk_properties:
            df[variable] = extract_bulk_parameter(variable, partition)
        bulk.append(df)
    return bulk


def worker(spectrum):
    return partition_spectrum(spectrum)


def partition_spectra(spectra2D: List[FrequencyDirectionSpectrum],
                      minimum_duration: timedelta,
                      config=None, indent='') -> List[
    List[FrequencyDirectionSpectrum]]:
    if config:
        for key in config:
            assert key in DEFAULT_CONFIG_PARTITION_SPECTRA, f"{key} is not a valid configuration entry"

        config = DEFAULT_CONFIG_PARTITION_SPECTRA | config
    else:
        config = DEFAULT_CONFIG_PARTITION_SPECTRA

    # Step 1: Partition the data
    logger.info( indent + ' - Partition Data')
    raw_partitions = []
    if config['parallel']:
        with get_context("spawn").Pool(processes=cpu_count()) as pool:
            output = pool.map(worker, spectra2D,
                              chunksize=len(spectra2D) // cpu_count())
        raw_partitions = [partition for partition, _ in output]
    else:
        for index, spectrum in enumerate(spectra2D):
            logger.debug(indent + f'\t {index:05d} out of {len(spectra2D)}')
            partitions, _ = partition_spectrum(spectrum,
                                               config['partitionConfig'])
            raw_partitions.append(partitions)

    # Step 2: create a graph
    logger.info( indent + ' - Create Graph')
    graph = create_graph(raw_partitions, minimum_duration)

    # Step 3: create wave field from the graph
    logger.info( indent + ' - Create Wave Fields From Graph')
    wave_fields = wave_fields_from(graph)

    # Step 4: Postprocessing

    # Apply a filter on the bulk parameters
    logger.info( indent + ' - Apply Bulk Filter')
    if config['fieldFiltersettings']['filter']:
        wave_fields = filter_fields(
            wave_fields,
            min_duration=minimum_duration,
            config=config['fieldFiltersettings']
        )
    return wave_fields


# -----------------------------------------------------------------------------
#                       Boilerplate Interfaces
# -----------------------------------------------------------------------------
@overload
def get_bulk_partitions_from_spectral_partitions(
        spectral_partitions: Dict[str, List[List[FrequencyDirectionSpectrum]]]
) -> Dict[str, List[DataFrame]]: ...


@overload
def get_bulk_partitions_from_spectral_partitions(
        spectral_partitions: List[List[FrequencyDirectionSpectrum]]
) -> List[DataFrame]: ...


def get_bulk_partitions_from_spectral_partitions(
        spectral_partitions: Union[Dict[str, List[List[
            FrequencyDirectionSpectrum]]], List[
            List[FrequencyDirectionSpectrum]]]) -> Union[Dict[
                                                 str, List[DataFrame]], List[
                                                 DataFrame]]:
    if isinstance(spectral_partitions, dict):
        output = {}
        for key in spectral_partitions:
            output[key] = bulk_parameters_partitions(spectral_partitions[key])

    elif isinstance(spectral_partitions, list):
        output = bulk_parameters_partitions(spectral_partitions)
    else:
        raise Exception('Cannot process input')

    return output


def get_spectral_partitions_from_2dspectra(
        spectra: Union[Dict[str, List[FrequencyDirectionSpectrum]],
                       List[FrequencyDirectionSpectrum]],
        minimum_duration: timedelta,
        config=None,
        verbose=False) -> Union[
    Dict[str, List[List[FrequencyDirectionSpectrum]]], List[List[
        FrequencyDirectionSpectrum]]]:
    #

    if isinstance(spectra, dict):
        output = {}
        for key, item in spectra.items():
            output[key] = partition_spectra(item,
                                            minimum_duration,
                                            config)
    elif isinstance(spectra, list):
        output = partition_spectra(spectra,
                                   minimum_duration,
                                   config)
    else:
        raise Exception('Cannot process input')

    return output



"""
Copyright (C) 2022
Sofar Ocean Technologies

Authors: Pieter Bart Smit
======================

Module defining the various resources we can use. Here a resource is a class
that acts as a normal file opened for binary read or write in Python. However,
the underlying object may be a an s3 object or a local file. In the case of
an s3 object all IO is buffered and changes occur on s3 only once the file
object is closed.

Classes:
- `MetaData`, dataclass containing metadata from a restart file

Functions:

- `read_header`, read the header information from a restart file opened by the
    given resource.


How To Use This Module
======================
(See the individual functions for details.)

1. import
2. read metadat using `read_header` which returns a MetaData object.
"""

from dataclasses import dataclass
from datetime import datetime, timezone
from io import BytesIO
from typing import Literal

from roguewave.wavewatch3.fortran_types import FortranCharacter, FortranInt
from roguewave.wavewatch3.resources import Resource


@dataclass()
class MetaData:
    """
    Class containing all the metadata we need that is stored in a ww3 restart
    file.
    """
    name: str
    version: "str"
    grid_name: "str"
    restart_type: "str"
    nsea: int
    nspec: int
    record_size_bytes: int
    time: datetime
    byte_order: Literal["<", ">", "="]
    float_size: int


def read_header(resource: Resource,
                number_of_spectral_points = 36 * 36,
                byte_order = '<', float_size = 4) -> MetaData:
    """
    Function to read header information in a restart file and return the
    needed MetaData object.

    :param resource: an instance of a resource.
    :param number_of_spectral_points: number of spectral points. This may not
        be known in advance- but due to the way the file is layed out a guess
        that is larger than 78 bytes will work
    :param byte_order: Sofar ww3 uses little endian
    :param float_size: 4 bytes for Sofar (float32)
    :return:
    """
    data = {}

    # first read the character arrays for name, version, grid name and
    # restart_type. We do not know the record size yet with which the file
    # was written, so we just guess a record size.
    guess_record_size_bytes = number_of_spectral_points * float_size
    stream = BytesIO(resource.read(guess_record_size_bytes * 2))

    fort_char = FortranCharacter(endianness=byte_order)
    fort_int = FortranInt(endianness=byte_order)

    data['byte_order'] = byte_order
    data['float_size'] = float_size
    data["name"] = fort_char.unpack(stream, 26)
    data["version"] = fort_char.unpack(stream, 10)
    data["grid_name"] = fort_char.unpack(stream, 30)
    data["restart_type"] = fort_char.unpack(stream, 4)

    # Now we can read the number of spatial points and number of spectral points
    data["nsea"] = fort_int.unpack(stream, 1)[0]
    data["nspec"] = fort_int.unpack(stream, 1)[0]
    data['record_size_bytes'] = \
        data["nspec"] * float_size

    if (guess_record_size_bytes*2 < data['record_size_bytes'] + 8):
        # We only need to read 8 more bytes than the actual length to get the
        # time information. Now we know the actual record length, see if we
        # can just go ahead, or if we need to reload data from the resource.
        # Especially if the restart file is remote this saves us another
        # request.
        stream.seek(0)
        stream =   BytesIO(resource.read(data['record_size_bytes'] + 8))

    # Jump to the second record which contains time information
    stream.seek(data['record_size_bytes'])

    date_int = fort_int.unpack(stream, 1)[0]
    year, month, day = _unpack_date_time_from_int(date_int)

    time_int = fort_int.unpack(stream, 1)[0]
    hour, min, sec = _unpack_date_time_from_int(time_int)

    data["time"] = datetime(year, month, day, hour, min, sec,
                            tzinfo=timezone.utc)
    return MetaData(**data)


def _unpack_date_time_from_int(t):
    """
    Wavewatch 3 restart files store dates and times as integers in the form:
        20220101 for '2022-01-01' or 193000 for '19:30:00Z'. This is a simple
        function to unpack either the year/month/day ore hour/minute/second
        triple.
    :param t: integer containing date or time
    :return: time or date tripple.
    """
    x = int(t / 10000)
    y = int((t - x * 10000) / 100)
    z = t - x * 10000 - 100 * y
    return x,y,z